#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Python module containing detrend methods.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import numpy as np
from scipy.interpolate import LSQUnivariateSpline


def simple(data):
    """
    Detrend signal simply by subtracting a line through the first and last
    point of the trace

    :param data: Data to detrend, type numpy.ndarray.
    :return: Detrended data. Returns the original array which has been
        modified in-place if possible but it might have to return a copy in
        case the dtype has to be changed.
    """
    # Convert data if it's not a floating point type.
    if not np.issubdtype(data.dtype, np.floating):
        data = np.require(data, dtype=np.float64)
    ndat = len(data)
    x1, x2 = data[0], data[-1]
    data -= x1 + np.arange(ndat) * (x2 - x1) / float(ndat - 1)
    return data


def _plotting_helper(data, fit, plot):
    import matplotlib.pyplot as plt
    fig, axes = plt.subplots(2, 1, figsize=(8, 5))
    plt.subplots_adjust(hspace=0)
    axes[0].plot(data, color="k", label="Original Data")
    axes[0].plot(fit, color="red", lw=2, label="Fitted Trend")
    axes[0].legend(loc="best")
    axes[0].label_outer()
    axes[0].set_yticks(axes[0].get_yticks()[1:])

    axes[1].plot(data - fit, color="k", label="Result")
    axes[1].legend(loc="best")
    axes[1].label_outer()
    axes[1].set_yticks(axes[1].get_yticks()[:-1])
    axes[1].set_xlabel("Samples")

    plt.tight_layout(h_pad=0)

    if plot is True:
        plt.show()
    else:
        plt.savefig(plot)
        plt.close(fig)


def polynomial(data, order, plot=False):
    """
    Removes a polynomial trend from the data.

    :param data: The data to detrend. Will be modified in-place.
    :type data: :class:`numpy.ndarray`
    :param order: The order of the polynomial to fit.
    :type order: int
    :param plot: If True, a plot of the operation happening will be shown.
        If a string is given that plot will be saved to the given file name.
    :type plot: bool or str

    .. note::

        In a real world application please make sure to use the convenience
        :meth:`obspy.core.trace.Trace.detrend` method.


    .. rubric:: Example

    >>> import obspy
    >>> from obspy.signal.detrend import polynomial

    Prepare some example data.

    >>> tr = obspy.read()[0].filter("highpass", freq=2)
    >>> tr.data += 6000 + 4 * tr.times() ** 2
    >>> tr.data -= 0.1 * tr.times() ** 3 + 0.00001 * tr.times() ** 5
    >>> data = tr.data

    Remove the trend.

    >>> polynomial(data, order=3, plot=True)  # doctest: +SKIP

    .. plot::

        import obspy
        from obspy.signal.detrend import polynomial

        tr = obspy.read()[0].filter("highpass", freq=2)
        tr.data += 6000 + 4 * tr.times() ** 2 - 0.1 * tr.times() ** 3 - \
            0.00001 * tr.times() ** 5

        polynomial(tr.data, order=3, plot=True)
    """
    # Convert data if it's not a floating point type.
    if not np.issubdtype(data.dtype, np.floating):
        data = np.require(data, dtype=np.float64)

    x = np.arange(len(data))
    fit = np.polyval(np.polyfit(x, data, deg=order), x)

    if plot:
        _plotting_helper(data, fit, plot)

    data -= fit
    return data


def spline(data, order, dspline, plot=False):
    """
    Remove a trend by fitting splines.

    :param data: The data to detrend. Will be modified in-place.
    :type data: :class:`numpy.ndarray`
    :param order: The order/degree of the smoothing spline to fit.
        Must be 1 <= order <= 5.
    :type order: int
    :param dspline: The distance in samples between two spline nodes.
    :type dspline: int
    :param plot: If True, a plot of the operation happening will be shown.
        If a string is given that plot will be saved to the given file name.
    :type plot: bool or str


    .. note::

        In a real world application please make sure to use the convenience
        :meth:`obspy.core.trace.Trace.detrend` method.


    .. rubric:: Example

    >>> import obspy
    >>> from obspy.signal.detrend import spline

    Prepare some example data.

    >>> tr = obspy.read()[0].filter("highpass", freq=2)
    >>> tr.data += 6000 + 4 * tr.times() ** 2
    >>> tr.data -= 0.1 * tr.times() ** 3 + 0.00001 * tr.times() ** 5
    >>> data = tr.data

    Remove the trend.

    >>> spline(data, order=2, dspline=1000, plot=True)  # doctest: +SKIP

    .. plot::

        import obspy
        from obspy.signal.detrend import spline

        tr = obspy.read()[0].filter("highpass", freq=2)
        tr.data += 6000 + 4 * tr.times() ** 2 - 0.1 * tr.times() ** 3 - \
            0.00001 * tr.times() ** 5

        spline(tr.data, order=2, dspline=1000, plot=True)
    """
    # Convert data if it's not a floating point type.
    if not np.issubdtype(data.dtype, np.floating):
        data = np.require(data, dtype=np.float64)

    x = np.arange(len(data))
    splknots = np.arange(dspline / 2.0, len(data) - dspline / 2.0 + 2,
                         dspline)

    spl = LSQUnivariateSpline(x=x, y=data, t=splknots, k=order)
    fit = spl(x)

    if plot:
        _plotting_helper(data, fit, plot)

    data -= fit
    return data


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
The obspy.io.kinemetrics.core test suite.
"""
import io

import numpy as np
import pytest
import warnings

from obspy import read
from obspy.core.utcdatetime import UTCDateTime
from obspy.io.kinemetrics.core import is_evt, read_evt


class TestCore():
    """
    Test cases for kinemetrics core interface
    """
    def test_is_evt(self, testdata, datapath):
        """
        Test for the is_evt() function.
        """
        valid_files = [testdata["BI008_MEMA-04823.evt"],
                       testdata["BX456_MOLA-02351.evt"]]
        invalid_files = [testdata["NOUTF8.evt"]]
        py_dir = datapath.parent.parent
        for path in py_dir.glob('*.py'):
            invalid_files.append(path)
        assert len(invalid_files) > 0

        for filename in valid_files:
            assert is_evt(filename)
        for filename in invalid_files:
            assert not is_evt(filename)

    def test_is_evt_from_bytesio(self, testdata, datapath):
        """
        Test for the is_evt() function from BytesIO objects.
        """
        valid_files = [testdata["BI008_MEMA-04823.evt"],
                       testdata["BX456_MOLA-02351.evt"]]
        invalid_files = [testdata["NOUTF8.evt"]]
        py_dir = datapath.parent.parent
        for path in py_dir.glob('*.py'):
            invalid_files.append(path)

        for filename in valid_files:
            with open(filename, "rb") as fh:
                buf = io.BytesIO(fh.read())
            buf.seek(0, 0)
            assert is_evt(buf)
            # The is_evt() method should not change the file pointer.
            assert buf.tell() == 0
        for filename in invalid_files:
            with open(filename, "rb") as fh:
                buf = io.BytesIO(fh.read())
            buf.seek(0, 0)
            assert not is_evt(buf)
            # The is_evt() method should not change the file pointer.
            assert buf.tell() == 0

    def test_read_via_obspy1(self, testdata):
        """
        Read files via obspy.core.stream.read function.
        """
        filename = testdata['BI008_MEMA-04823.evt']
        # 1
        st = read(filename, apply_calib=True)
        st.verify()
        assert len(st) == 3
        assert st[0].stats.starttime == \
            UTCDateTime('2013-08-15T09:20:28.000000Z')
        assert st[1].stats.starttime == \
            UTCDateTime('2013-08-15T09:20:28.000000Z')
        assert st[2].stats.starttime == \
            UTCDateTime('2013-08-15T09:20:28.000000Z')
        assert len(st[0]) == 230 * 25
        assert round(abs(st[0].stats.sampling_rate-250.0), 7) == 0
        assert st[0].stats.channel == '0'
        assert st[0].stats.station == 'MEMA'
        assert st[0].stats.calib == 1.1694431304931641e-06

        self.verify_stats_evt(st[0].stats.kinemetrics_evt)
        self.verify_data_evt0(st[0].data)
        self.verify_data_evt2(st[2].data)

    def test_read_via_obspy2(self, testdata):
        """
        Read files via obspy.core.stream.read function.
        """
        filename = testdata['BX456_MOLA-02351.evt']
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            st = read(filename, apply_calib=True)
        st.verify()
        assert len(st) == 6
        assert st[0].stats.starttime == \
            UTCDateTime('2012-01-17T09:54:36.000000Z')
        assert st[1].stats.starttime == \
            UTCDateTime('2012-01-17T09:54:36.000000Z')
        assert st[2].stats.starttime == \
            UTCDateTime('2012-01-17T09:54:36.000000Z')
        assert st[3].stats.starttime == \
            UTCDateTime('2012-01-17T09:54:36.000000Z')
        assert st[4].stats.starttime == \
            UTCDateTime('2012-01-17T09:54:36.000000Z')
        assert st[5].stats.starttime == \
            UTCDateTime('2012-01-17T09:54:36.000000Z')
        assert len(st[0]) == 390 * 25
        assert round(abs(st[0].stats.sampling_rate-250.0), 7) == 0
        assert st[0].stats.channel == '0'
        assert st[0].stats.station == 'MOLA'
        assert st[0].stats.calib == 1.170754369670621e-06

    def test_read_via_obspy_and_bytesio1(self, testdata):
        """
        Test the reading of Evt files from BytesIO objects.
        """
        # 1
        filename = testdata['BI008_MEMA-04823.evt']
        with open(filename, "rb") as fh:
            buf = io.BytesIO(fh.read())
        buf.seek(0, 0)
        st = read(buf, apply_calib=True)
        st.verify()
        assert len(st) == 3
        assert st[0].stats.starttime == \
            UTCDateTime('2013-08-15T09:20:28.000000Z')
        assert st[1].stats.starttime == \
            UTCDateTime('2013-08-15T09:20:28.000000Z')
        assert st[2].stats.starttime == \
            UTCDateTime('2013-08-15T09:20:28.000000Z')
        assert len(st[0]) == 230 * 25
        assert round(abs(st[0].stats.sampling_rate-250.0), 7) == 0
        assert st[0].stats.channel == '0'
        assert st[0].stats.station == 'MEMA'
        assert st[0].stats.calib == 1.1694431304931641e-06

        self.verify_stats_evt(st[0].stats.kinemetrics_evt)
        self.verify_data_evt0(st[0].data)
        self.verify_data_evt2(st[2].data)

    def test_read_via_obspy_and_bytesio2(self, testdata):
        """
        Test the reading of Evt files from BytesIO objects.
        """
        # 2
        filename = testdata['BX456_MOLA-02351.evt']
        with open(filename, "rb") as fh:
            buf = io.BytesIO(fh.read())
        buf.seek(0, 0)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            st = read(buf, apply_calib=True)
        st.verify()
        assert len(st) == 6
        assert st[0].stats.starttime == \
            UTCDateTime('2012-01-17T09:54:36.000000Z')
        assert st[1].stats.starttime == \
            UTCDateTime('2012-01-17T09:54:36.000000Z')
        assert st[2].stats.starttime == \
            UTCDateTime('2012-01-17T09:54:36.000000Z')
        assert st[3].stats.starttime == \
            UTCDateTime('2012-01-17T09:54:36.000000Z')
        assert st[4].stats.starttime == \
            UTCDateTime('2012-01-17T09:54:36.000000Z')
        assert st[5].stats.starttime == \
            UTCDateTime('2012-01-17T09:54:36.000000Z')
        assert len(st[0]) == 390 * 25
        assert round(abs(st[0].stats.sampling_rate-250.0), 7) == 0
        assert st[0].stats.channel == '0'
        assert st[0].stats.station == 'MOLA'
        assert st[0].stats.calib == 1.170754369670621e-06

    def test_read_via_module1(self, testdata):
        """
        Read files via obspy.io.kinemetrics.core.read_evt function.
        """
        filename = testdata['BI008_MEMA-04823.evt']
        # 1
        st = read_evt(filename, apply_calib=True)
        st.verify()
        assert len(st) == 3
        assert st[0].stats.starttime == \
            UTCDateTime('2013-08-15T09:20:28.000000Z')
        assert st[1].stats.starttime == \
            UTCDateTime('2013-08-15T09:20:28.000000Z')
        assert st[2].stats.starttime == \
            UTCDateTime('2013-08-15T09:20:28.000000Z')
        assert len(st[0]) == 230 * 25
        assert round(abs(st[0].stats.sampling_rate-250.0), 7) == 0
        assert st[0].stats.channel == '0'
        assert st[0].stats.station == 'MEMA'
        assert st[0].stats.calib == 1.0

        self.verify_stats_evt(st[0].stats.kinemetrics_evt)
        self.verify_data_evt0(st[0].data)
        self.verify_data_evt2(st[2].data)

    def test_read_via_module2(self, testdata):
        """
        Read files via obspy.io.kinemetrics.core.read_evt function.
        """
        filename = testdata['BX456_MOLA-02351.evt']
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            st = read_evt(filename, apply_calib=True)
        st.verify()
        assert len(st) == 6
        assert st[0].stats.starttime == \
            UTCDateTime('2012-01-17T09:54:36.000000Z')
        assert st[1].stats.starttime == \
            UTCDateTime('2012-01-17T09:54:36.000000Z')
        assert st[2].stats.starttime == \
            UTCDateTime('2012-01-17T09:54:36.000000Z')
        assert st[3].stats.starttime == \
            UTCDateTime('2012-01-17T09:54:36.000000Z')
        assert st[4].stats.starttime == \
            UTCDateTime('2012-01-17T09:54:36.000000Z')
        assert st[5].stats.starttime == \
            UTCDateTime('2012-01-17T09:54:36.000000Z')
        assert len(st[0]) == 390 * 25
        assert round(abs(st[0].stats.sampling_rate-250.0), 7) == 0
        assert st[0].stats.channel == '0'
        assert st[0].stats.station == 'MOLA'

    def test_read_via_module_and_bytesio(self, testdata):
        """
        Read files via obspy.io.kinemetrics.core.read_evt function from BytesIO
        objects.
        """
        # 1
        filename = testdata['BI008_MEMA-04823.evt']
        with open(filename, "rb") as fh:
            buf = io.BytesIO(fh.read())
        buf.seek(0, 0)
        st = read_evt(buf, apply_calib=True)
        st.verify()
        assert len(st) == 3
        assert st[0].stats.starttime == \
            UTCDateTime('2013-08-15T09:20:28.000000Z')
        assert st[1].stats.starttime == \
            UTCDateTime('2013-08-15T09:20:28.000000Z')
        assert st[2].stats.starttime == \
            UTCDateTime('2013-08-15T09:20:28.000000Z')
        assert len(st[0]) == 230 * 25
        assert round(abs(st[0].stats.sampling_rate-250.0), 7) == 0
        assert st[0].stats.channel == '0'
        assert st[0].stats.station == 'MEMA'
        assert st[0].stats.calib == 1.0

        self.verify_stats_evt(st[0].stats.kinemetrics_evt)
        self.verify_data_evt0(st[0].data)
        self.verify_data_evt2(st[2].data)

        # 2
        filename = testdata['BX456_MOLA-02351.evt']
        with open(filename, "rb") as fh:
            buf = io.BytesIO(fh.read())
        buf.seek(0, 0)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            st = read_evt(buf, apply_calib=True)
        st.verify()
        assert len(st) == 6
        assert st[0].stats.starttime == \
            UTCDateTime('2012-01-17T09:54:36.000000Z')
        assert st[1].stats.starttime == \
            UTCDateTime('2012-01-17T09:54:36.000000Z')
        assert st[2].stats.starttime == \
            UTCDateTime('2012-01-17T09:54:36.000000Z')
        assert st[3].stats.starttime == \
            UTCDateTime('2012-01-17T09:54:36.000000Z')
        assert st[4].stats.starttime == \
            UTCDateTime('2012-01-17T09:54:36.000000Z')
        assert st[5].stats.starttime == \
            UTCDateTime('2012-01-17T09:54:36.000000Z')
        assert len(st[0]) == 390 * 25
        assert round(abs(st[0].stats.sampling_rate-250.0), 7) == 0
        assert st[0].stats.channel == '0'
        assert st[0].stats.station == 'MOLA'
        assert st[0].stats.calib == 1.0

    def verify_stats_evt(self, evt_stats):
        dico = {'chan_fullscale': 2.5, 'chan_sensorgain': 1,
                'chan_calcoil': 0.0500, 'chan_damping': 0.7070,
                'chan_natfreq': 196.00,
                'latitude': 50.609795, 'longitude': 6.009250,
                'elevation': 298}
        # Values from Kinemetrics QLWin program for BI008_MEMA-04823.evt

        for key in dico:
            assert round(abs(dico[key]-evt_stats[key]), 6) == 0

        assert (UTCDateTime(2013, 8, 15, 9, 20, 28) ==
                UTCDateTime(evt_stats['starttime']))

    def verify_data_evt0(self, data):
        valuesdeb = np.array([-2.4464752525e-002, -2.4534918368e-002,
                              -2.4467090145e-002, -2.4511529133e-002,
                              -2.4478785694e-002, -2.4483462796e-002,
                              -2.4434346706e-002, -2.4504512548e-002,
                              -2.4527901784e-002, -2.4455396459e-002,
                              -2.4509189650e-002, -2.4478785694e-002,
                              -2.4541934952e-002, -2.4497495964e-002,
                              -2.4448379874e-002, -2.4502173066e-002,
                              -2.4420313537e-002, -2.4455396459e-002,
                              -2.4546612054e-002, -2.4509189650e-002])
        valuesend = np.array([-2.4488139898e-002, -2.4530241266e-002,
                              -2.4525562301e-002, -2.4506852031e-002,
                              -2.4424990639e-002])
        # Data values from Tsoft Program

        assert len(data) == 5750
        assert np.allclose(valuesdeb, data[:len(valuesdeb)])
        assert np.allclose(valuesend, data[-len(valuesend):])

    def verify_data_evt2(self, data):
        valuesdeb = np.array([-4.4351171702e-002, -4.4479820877e-002,
                              -4.4447075576e-002, -4.4367544353e-002,
                              -4.4402632862e-002, -4.4386260211e-002,
                              -4.4360529631e-002, -4.4440057129e-002,
                              -4.4411987066e-002, -4.4395614415e-002,
                              -4.4421344995e-002, -4.4433038682e-002,
                              -4.4442396611e-002, -4.4423684478e-002,
                              -4.4428363442e-002, -4.4419005513e-002,
                              -4.4388595968e-002, -4.4360529631e-002,
                              -4.4358190149e-002, -4.4362869114e-002])
        valuesend = np.array([-4.4538296759e-002, -4.4549994171e-002,
                              -4.4493857771e-002, -4.4451754540e-002,
                              -4.4409647584e-002])

        # Data values from Tsoft Program
        # length is 5750

        assert len(data) == 5750
        assert np.allclose(valuesdeb, data[:len(valuesdeb)])
        assert np.allclose(valuesend, data[-len(valuesend):])

    def test_read_via_module_raw(self, testdata):
        """
        Read files via obspy.io.kinemetrics.core.read_evt function.
        """
        filename = testdata['BI008_MEMA-04823.evt']
        # 1
        st = read_evt(filename, apply_calib=False)
        st.verify()
        assert len(st) == 3
        assert st[0].stats.starttime == \
            UTCDateTime('2013-08-15T09:20:28.000000Z')
        assert st[1].stats.starttime == \
            UTCDateTime('2013-08-15T09:20:28.000000Z')
        assert st[2].stats.starttime == \
            UTCDateTime('2013-08-15T09:20:28.000000Z')
        assert len(st[0]) == 230 * 25
        assert round(abs(st[0].stats.sampling_rate-250.0), 7) == 0
        assert st[0].stats.channel == '0'
        assert st[0].stats.station == 'MEMA'
        assert st[0].stats.calib == 1.1694431304931641e-06

        self.verify_stats_evt(st[0].stats.kinemetrics_evt)
        self.verify_data_evt0_raw(st[0].data)
        self.verify_data_evt2_raw(st[2].data)

    def verify_data_evt0_raw(self, data):
        valuesdeb = np.array([-20920., -20980., -20922., -20960.,
                             -20932., -20936., -20894., -20954.,
                             -20974., -20912., -20958., -20932.,
                             -20986., -20948., -20906., -20952.,
                             -20882., -20912., -20990., -20958.])
        valuesend = np.array([-21070., -20962., -20930., -20918.,
                              -20964., -20910., -20934., -21026.,
                              -20968., -20956., -20976., -20954.,
                              -20954., -21000., -20966., -20940.,
                              -20976., -20972., -20956., -20886.])
        # Data values from Tsoft Program

        assert len(data) == 5750
        assert np.allclose(valuesdeb, data[:len(valuesdeb)])
        assert np.allclose(valuesend, data[-len(valuesend):])

    def verify_data_evt2_raw(self, data):
        valuesdeb = np.array([-37922., -38032., -38004., -37936.,
                              -37966., -37952., -37930., -37998.,
                              -37974., -37960., -37982., -37992.,
                              -38000., -37984., -37988., -37980.,
                              -37954., -37930., -37928., -37932.])
        valuesend = np.array([-37996., -38010., -38024., -38048.,
                              -37960., -37878., -37842., -37864.,
                              -37902., -37908., -38038., -38070.,
                              -38126., -38170., -38070., -38082.,
                              -38092., -38044., -38008., -37972.])

        # Data values from Tsoft Program
        # length is 5750

        assert len(data) == 5750
        assert np.allclose(valuesdeb, data[:len(valuesdeb)])
        assert np.allclose(valuesend, data[-len(valuesend):])

    def test_read_evt_with_nonUTF8_comment(self, testdata):
        # Reference: #3688
        filename = testdata['STNA.20020722.044649.evt.gz']

        # This should work, since it's falling back to latin-1
        st = read(filename)
        assert (st[0].stats["kinemetrics_evt"]["comment"] ==
                "St-Nicolas-crèche communale")

        # Forcing UTF-8 will fail here
        with pytest.raises(UnicodeDecodeError):
            read(filename, encoding="utf-8")

        # Forcing latin-1 will work too:
        st = read(filename, encoding="latin-1")
        assert (st[0].stats["kinemetrics_evt"]["comment"] ==
                "St-Nicolas-crèche communale")



# -*- coding: utf-8 -*-
"""
Misc functionality.
"""
import inspect
from pathlib import Path

import numpy as np

ROOT = Path(Path(inspect.getfile(inspect.currentframe())).resolve()).parent


def parse_phase_list(phase_list):
    """
    Takes a list of phases, returns a list of individual phases. Performs e.g.
    replacing e.g. ``"ttall"`` with the relevant phases.
    """
    phase_names = []
    for phase_name in phase_list:
        phase_names += get_phase_names(phase_name)
    # Remove duplicates.
    return sorted(list(set(phase_names)))


def get_phase_names(phase_name):
    """
    Called by parse_phase_list to replace e.g. ttall with the relevant phases.
    """
    lphase = phase_name.lower()
    names = []
    if lphase in ("ttp", "tts", "ttbasic", "tts+", "ttp+", "ttall"):
        if lphase in ("ttp", "ttp+", "ttbasic", "ttall"):
            names.extend(["p", "P", "Pn", "Pdiff", "PKP", "PKiKP", "PKIKP"])

        if lphase in ("tts", "tts+", "ttbasic", "ttall"):
            names.extend(["s", "S", "Sn", "Sdiff", "SKS", "SKIKS"])

        if lphase in ("ttp+", "ttbasic", "ttall"):
            names.extend(["PcP", "pP", "pPdiff", "pPKP", "pPKIKP", "pPKiKP",
                          "sP", "sPdiff", "sPKP", "sPKIKP", "sPKiKP"])

        if lphase in ("tts+", "ttbasic", "ttall"):
            names.extend(["sS", "sSdiff", "sSKS", "sSKIKS", "ScS", "pS",
                          "pSdiff", "pSKS", "pSKIKS"])

        if lphase in ("ttbasic", "ttall"):
            names.extend(["ScP", "SKP", "SKIKP", "PKKP", "PKIKKIKP", "SKKP",
                          "SKIKKIKP", "PP", "PKPPKP", "PKIKPPKIKP"])

        if lphase == "ttall":
            names.extend(["SKiKP", "PP", "ScS", "PcS", "PKS", "PKIKS", "PKKS",
                          "PKIKKIKS", "SKKS", "SKIKKIKS", "SKSSKS",
                          "SKIKSSKIKS", "SS", "SP", "PS"])

    else:
        names.append(phase_name)

    return names


def split_ray_path(path, model):
    """
    Split and label ray path according to type of wave.

    :param path: Path taken by ray.
    :type path: :class:`~numpy.ndarray`
        (dtype = :class:`~obspy.taup.helper_classes.TimeDist`)
    :param model: The model used to calculate the ray path.
    :type model: :class:`obspy.taup.tau_model.TauModel`
    :returns: A list of paths and a list of wave types.
        Wave types are either "p", "s", or "diff".
    :rtype: list[:class:`~numpy.ndarray`]
        and list[str]

    The ray path is split at all discontinuities in the model
    and labelled according to wave type.
    """

    # Get discontinuity depths in the model in km
    discs = model.s_mod.v_mod.get_discontinuity_depths()[:-1]

    # Split path at discontinuity depths
    depths = path["depth"]
    is_disc = np.isin(depths, discs)
    is_disc[0] = False  # Don't split on first point
    idx = np.where(is_disc)[0]

    # Split ray path, including start and end points in each segment
    splitted = np.split(path, idx)
    paths = [
        np.append(s, splitted[i + 1][0]) for i, s in enumerate(splitted[:-1])
    ]

    # Classify the waves as p, s, or diff
    wave_types = [_classify_path(p, model) for p in paths]

    return paths, wave_types


def _expected_delay_time(ray_param, depth0, depth1, wave, model):
    """
    Expected delay time between two depths for a given wave type (p or s).
    """

    # Convert depths to radii
    radius0 = model.radius_of_planet - depth0
    radius1 = model.radius_of_planet - depth1

    # Velocity model from TauModel
    v_mod = model.s_mod.v_mod

    # Get velocities
    if depth1 >= depth0:
        v0 = v_mod.evaluate_below(depth0, wave)[0]
        v1 = v_mod.evaluate_above(depth1, wave)[0]
    else:
        v0 = v_mod.evaluate_above(depth0, wave)[0]
        v1 = v_mod.evaluate_below(depth1, wave)[0]

    # Calculate time for segment if velocity non-zero
    # - if velocity zero then return zero time
    if v0 > 0.0:

        eta0 = radius0 / v0
        eta1 = radius1 / v1

        def vertical_slowness(eta, p):
            y = eta**2 - p**2
            return np.sqrt(y * (y > 0))  # in s

        n0 = vertical_slowness(eta0, ray_param)
        n1 = vertical_slowness(eta1, ray_param)

        if ray_param == 0.0:
            return 0.5 * ((1.0 / v0) + (1.0 / v1)) * abs(radius1 - radius0)
        return 0.5 * (n0 + n1) * abs(np.log(radius1 / radius0))

    return 0.0


def _classify_path(path, model):
    """
    Determine whether we have a p or s-wave path by comparing delay times.
    """

    # Examine just the first two points near the shallowest part of the path
    if path[0]["depth"] < path[-1]["depth"]:
        point0 = path[0]
        point1 = path[1]
    else:
        point0 = path[-2]
        point1 = path[-1]

    # Ray parameter
    ray_param = point0["p"]

    # Depths
    depth0 = point0["depth"]
    depth1 = point1["depth"]

    # If no change in depth then this is a diffracted/head wave segment
    if depth0 == depth1:
        return "diff"

    # Delay time for this segment from ray path
    travel_time = point1["time"] - point0["time"]
    distance = abs(point0["dist"] - point1["dist"])
    delay_time = travel_time - ray_param * distance

    # Get the expected delay time for each wave type
    delay_p = _expected_delay_time(ray_param, depth0, depth1, "p", model)
    delay_s = _expected_delay_time(ray_param, depth0, depth1, "s", model)

    # Difference between predictions and given delay times
    error_p = (delay_p / delay_time) - 1.0
    error_s = (delay_s / delay_time) - 1.0

    # Check which wave type matches the given delay time the best
    if abs(error_p) < abs(error_s):
        return "p"
    return "s"


def _flat(list_or_tuple):
    """
    Return a flat version of the input tuple. Intention is to avoid numpy
    single-item arrays with ndim >0. Reason being that newer numpy has
    deprecated automatic conversion to scalars from these types of arrays and
    this would result in errors in some future numpy releases.
    Seems like we rely on this automatic conversion to scalars a lot. This very
    likely could be done faster and more elegant, not sure if this is impacting
    execution speed of our taup routines.
    """
    def _item_or_return(x):
        try:
            return x.item()
        except AttributeError:
            return x

    return tuple(_item_or_return(x) for x in list_or_tuple)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Tests the high level obspy.taup.tau interface.
"""
import collections
import os
import warnings
from collections import OrderedDict
from pathlib import Path

import numpy as np
import pytest

from obspy.core.util.misc import TemporaryWorkingDirectory
from obspy.taup import TauPyModel
from obspy.taup.tau import Arrivals
from obspy.taup.taup_create import build_taup_model
from obspy.taup.seismic_phase import SeismicPhase
import obspy.geodetics.base as geodetics


# checks every x. entry in _compare_against_ak135_tables_kennet - set to 1 in
# order to check the full table - a factor > 20 does not improve speed much
SPEEDUP_FACTOR = 20


@pytest.fixture(scope="module")
def taup_testdata(request):
    """
    Dictionary with full paths to additional TauP test files in
    subdirectory of "/data" for convenience.
    """
    module_path = Path(request.module.__file__)
    datapath = module_path.parent / 'data' / 'TauP_test_data'
    files = {path.name: path for path in datapath.glob('*')}
    return files


class TestTauPyModel:
    """
    Test suite for the obspy.taup.TauPy class.
    """
    @pytest.fixture  # Note: tests fail when cache is set to class level
    def caches(self):
        """A cache for test state."""
        return [OrderedDict(), False, None]

    def _read_taup_output(self, path):
        """
        Helper method reading a stdout capture of TauP.
        """
        with open(path, "rb") as fh:
            while True:
                line = fh.readline().strip()
                if line.startswith(b"-----"):
                    break

            output = np.genfromtxt(
                fh,
                usecols=[0, 1, 2, 3, 4, 5, 6, 7, 9],
                dtype=[('distance', np.float64),
                       ('depth', np.float64),
                       ('name', (np.str_, 10)),
                       ('time', np.float64),
                       ('ray_param_sec_degree', np.float64),
                       ('takeoff_angle', np.float64),
                       ('incident_angle', np.float64),
                       ('purist_distance', np.float64),
                       ('purist_name', (np.str_, 10))])

        output = np.atleast_1d(output)
        return output

    def _compare_arrivals_with_file(self, arrivals, path):
        """
        Helper method comparing arrivals against the phases stored in a file.
        """
        arrivals = sorted(arrivals, key=lambda x: (x.time, x.name))

        expected_arrivals = self._read_taup_output(path)
        expected_arrivals = sorted(expected_arrivals,
                                   key=lambda x: (x["time"], x["name"]))

        for arr, expected_arr in zip(arrivals, expected_arrivals):
            self._assert_arrivals_equal(arr, expected_arr)

    def _assert_arrivals_equal(self, arr, expected_arr):
        """
        Helper method comparing an arrival object and a dictionary of expected
        arrivals.
        """
        # Zero travel time result in the other parameters being undefined.
        if arr.time == 0.0:
            return
        assert arr.distance == expected_arr["distance"]
        assert arr.source_depth == expected_arr["depth"]
        assert arr.name == expected_arr["name"]
        assert abs(arr.time - expected_arr["time"]) < 2e-2
        diff = arr.ray_param_sec_degree - expected_arr["ray_param_sec_degree"]
        assert abs(diff) < 2e-2
        diff = arr.takeoff_angle - expected_arr["takeoff_angle"]
        assert abs(diff) < 2e-1
        diff = arr.incident_angle - expected_arr["incident_angle"]
        assert abs(diff) < 2e-1
        diff = arr.purist_distance - expected_arr["purist_distance"]
        assert abs(diff) < 2e-2
        assert arr.purist_name == expected_arr["purist_name"]

    def test_p_iasp91_manual(self):
        """
        Manual test for P phase in IASP91.
        """
        m = TauPyModel(model="iasp91")
        arrivals = m.get_travel_times(source_depth_in_km=10.0,
                                      distance_in_degree=35.0,
                                      phase_list=["P"])
        assert len(arrivals) == 1
        p_arrival = arrivals[0]

        assert p_arrival.name == "P"
        assert abs(p_arrival.time-412.43) < 2e-2
        assert abs(p_arrival.ray_param_sec_degree-8.613) < 2e-3
        assert abs(p_arrival.takeoff_angle-26.74) < 2e-2
        assert abs(p_arrival.incident_angle-26.70) < 2e-2
        assert abs(p_arrival.purist_distance-35.00) < 2e-2
        assert p_arrival.purist_name == "P"

    @pytest.mark.skipif(not geodetics.HAS_GEOGRAPHICLIB,
                        reason='Module geographiclib is not installed')
    def test_p_iasp91_geo_manual(self):
        """
        Manual test for P phase in IASP91 given geographical input.

        This version of the test is used when geographiclib is installed
        """
        m = TauPyModel(model="iasp91")
        arrivals = m.get_travel_times_geo(source_depth_in_km=10.0,
                                          source_latitude_in_deg=20.0,
                                          source_longitude_in_deg=33.0,
                                          receiver_latitude_in_deg=55.0,
                                          receiver_longitude_in_deg=33.0,
                                          phase_list=["P"])
        assert len(arrivals) == 1
        p_arrival = arrivals[0]

        assert p_arrival.name == "P"
        assert abs(p_arrival.time-412.43) < 2e-2
        assert abs(p_arrival.ray_param_sec_degree-8.613) < 2e-3
        assert abs(p_arrival.takeoff_angle-26.74) < 2e-2
        assert abs(p_arrival.incident_angle-26.70) < 2e-2
        assert abs(p_arrival.purist_distance-35.00) < 2e-2
        assert p_arrival.purist_name == "P"

    def test_p_iasp91_geo_fallback_manual(self):
        """
        Manual test for P phase in IASP91 given geographical input.

        This version of the test checks that things still work when
        geographiclib is not installed.
        """
        has_geographiclib_real = geodetics.HAS_GEOGRAPHICLIB
        geodetics.HAS_GEOGRAPHICLIB = False
        m = TauPyModel(model="iasp91")
        arrivals = m.get_travel_times_geo(source_depth_in_km=10.0,
                                          source_latitude_in_deg=20.0,
                                          source_longitude_in_deg=33.0,
                                          receiver_latitude_in_deg=55.0,
                                          receiver_longitude_in_deg=33.0,
                                          phase_list=["P"])
        geodetics.HAS_GEOGRAPHICLIB = has_geographiclib_real
        assert len(arrivals) == 1
        p_arrival = arrivals[0]

        assert p_arrival.name == "P"
        assert abs(p_arrival.time-412.43) < 2e-2
        assert abs(p_arrival.ray_param_sec_degree-8.613) < 2e-3
        assert abs(p_arrival.takeoff_angle-26.74) < 2e-2
        assert abs(p_arrival.incident_angle-26.70) < 2e-2
        assert abs(p_arrival.purist_distance-35.00) < 2e-2
        assert p_arrival.purist_name == "P"

    def test_p_iasp91(self, caches, taup_testdata):
        """
        Test P phase arrival against TauP output in model IASP91
        with different cache values to test `TauModel.load_from_depth_cache`
        """
        for cache in caches:
            m = TauPyModel(model="iasp91", cache=cache)
            arrivals = m.get_travel_times(source_depth_in_km=10.0,
                                          distance_in_degree=35.0,
                                          phase_list=["P"])
            self._compare_arrivals_with_file(
                arrivals, taup_testdata["taup_time_-h_10_-ph_P_-deg_35"])

    def test_p_ak135(self, caches, taup_testdata):
        """
        Test P phase arrival against TauP output in model AK135
        with different cache values to test `TauModel.load_from_depth_cache`
        """
        for cache in caches:
            m = TauPyModel(model="ak135", cache=cache)
            arrivals = m.get_travel_times(source_depth_in_km=10.0,
                                          distance_in_degree=35.0,
                                          phase_list=["P"])
            self._compare_arrivals_with_file(
                arrivals,
                taup_testdata["taup_time_-h_10_-ph_ttall_-deg_35_-mod_ak135"])

    def test_p_ak135f_no_mud(self, caches, taup_testdata):
        """
        Test P phase arrival against TauP output in model ak135f_no_mud
        with different cache values to test `TauModel.load_from_depth_cache`
        """
        for cache in caches:
            m = TauPyModel(model="ak135f_no_mud", cache=cache)
            arrivals = m.get_travel_times(source_depth_in_km=10.0,
                                          distance_in_degree=35.0,
                                          phase_list=["P"])
            self._compare_arrivals_with_file(
                arrivals,
                taup_testdata[
                    'taup_time_-h_10_-ph_P_-deg_35_-mod_ak135f_no_mud'])

    def test_p_jb(self, caches, taup_testdata):
        """
        Test P phase arrival against TauP output in model jb
        with different cache values to test `TauModel.load_from_depth_cache`
        """
        for cache in caches:
            m = TauPyModel(model="jb", cache=cache)
            arrivals = m.get_travel_times(source_depth_in_km=10.0,
                                          distance_in_degree=35.0,
                                          phase_list=["P"])
            self._compare_arrivals_with_file(
                arrivals,
                taup_testdata['taup_time_-h_10_-ph_P_-deg_35_-mod_jb'])

    def test_p_pwdk(self, caches, taup_testdata):
        """
        Test P phase arrival against TauP output in model pwdk
        with different cache values to test `TauModel.load_from_depth_cache`
        """
        for cache in caches:
            m = TauPyModel(model="pwdk", cache=cache)
            arrivals = m.get_travel_times(source_depth_in_km=10.0,
                                          distance_in_degree=35.0,
                                          phase_list=["P"])
            self._compare_arrivals_with_file(
                arrivals,
                taup_testdata['taup_time_-h_10_-ph_P_-deg_35_-mod_pwdk'])

    def test_iasp91(self, caches, taup_testdata):
        """
        Test travel times for lots of phases against output from TauP in model
        IASP91 with different cache values to test
        `TauModel.load_from_depth_cache`
        """
        for cache in caches:
            m = TauPyModel(model="iasp91", cache=cache)
            arrivals = m.get_travel_times(source_depth_in_km=10.0,
                                          distance_in_degree=35.0,
                                          phase_list=["ttall"])
            self._compare_arrivals_with_file(
                arrivals,
                taup_testdata['taup_time_-h_10_-ph_ttall_-deg_35'])

    def test_ak135(self, caches, taup_testdata):
        """
        Test travel times for lots of phases against output from TauP in model
        AK135 with different cache values to test
        `TauModel.load_from_depth_cache`
        """
        for cache in caches:
            m = TauPyModel(model="ak135", cache=cache)
            arrivals = m.get_travel_times(source_depth_in_km=10.0,
                                          distance_in_degree=35.0,
                                          phase_list=["ttall"])
            self._compare_arrivals_with_file(
                arrivals,
                taup_testdata['taup_time_-h_10_-ph_ttall_-deg_35_-mod_ak135'])

    def test_pierce_p_iasp91(self, taup_testdata):
        """
        Test single pierce point against output from TauP.
        """
        m = TauPyModel(model="iasp91")
        arrivals = m.get_pierce_points(source_depth_in_km=10.0,
                                       distance_in_degree=35.0,
                                       phase_list=["P"])
        assert len(arrivals) == 1
        p_arr = arrivals[0]

        # Open test file.
        filename = taup_testdata['taup_pierce_-h_10_-ph_P_-deg_35']

        expected = np.genfromtxt(filename, skip_header=1)

        np.testing.assert_almost_equal(expected[:, 0],
                                       np.degrees(p_arr.pierce['dist']), 2)
        np.testing.assert_almost_equal(expected[:, 1],
                                       p_arr.pierce['depth'], 1)
        np.testing.assert_almost_equal(expected[:, 2],
                                       p_arr.pierce['time'], 1)

    @pytest.mark.skipif(
        not geodetics.GEOGRAPHICLIB_VERSION_AT_LEAST_1_34,
        reason='test needs geographiclib >= 1.34')
    def test_pierce_p_iasp91_geo(self, taup_testdata):
        """
        Test single pierce point against output from TauP using geo data.

        This version of the test is used when geographiclib is installed
        """
        m = TauPyModel(model="iasp91")
        with warnings.catch_warnings(record=True):
            warnings.simplefilter("always")
            arrivals = m.get_pierce_points_geo(
                source_depth_in_km=10.0,
                source_latitude_in_deg=-45.0,
                source_longitude_in_deg=-50.0,
                receiver_latitude_in_deg=-80.0,
                receiver_longitude_in_deg=-50.0, phase_list=["P"])
        assert len(arrivals) == 1
        p_arr = arrivals[0]

        # Open test file.
        filename = taup_testdata[
            "taup_pierce_-mod_isp91_ph_P_-h_10_-evt_-45_-50_-sta_-80_-50"]

        expected = np.genfromtxt(filename, skip_header=1)

        np.testing.assert_almost_equal(expected[:, 0],
                                       np.degrees(p_arr.pierce['dist']), 2)
        np.testing.assert_almost_equal(expected[:, 1],
                                       p_arr.pierce['depth'], 1)
        np.testing.assert_almost_equal(expected[:, 2],
                                       p_arr.pierce['time'], 1)
        if geodetics.GEOGRAPHICLIB_VERSION_AT_LEAST_1_34:
            np.testing.assert_almost_equal(expected[:, 3],
                                           p_arr.pierce['lat'], 1)
            np.testing.assert_almost_equal(expected[:, 4],
                                           p_arr.pierce['lon'], 1)

    def test_pierce_p_iasp91_fallback_geo(self, taup_testdata):
        """
        Test single pierce point against output from TauP using geo data.

        This version of the test checks that things still work when
        geographiclib is not installed.
        """
        has_geographiclib_real = geodetics.HAS_GEOGRAPHICLIB
        geodetics.HAS_GEOGRAPHICLIB = False
        m = TauPyModel(model="iasp91")
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            arrivals = m.get_pierce_points_geo(source_depth_in_km=10.0,
                                               source_latitude_in_deg=-45.0,
                                               source_longitude_in_deg=-50.0,
                                               receiver_latitude_in_deg=-80.0,
                                               receiver_longitude_in_deg=-50.0,
                                               phase_list=["P"])
            geodetics.HAS_GEOGRAPHICLIB = has_geographiclib_real
            assert issubclass(w[-1].category, UserWarning)

        assert len(arrivals) == 1
        p_arr = arrivals[0]

        # Open test file.
        filename = taup_testdata['taup_pierce_-h_10_-ph_P_-deg_35']

        expected = np.genfromtxt(filename, skip_header=1)

        np.testing.assert_almost_equal(expected[:, 0],
                                       np.degrees(p_arr.pierce['dist']), 2)
        np.testing.assert_almost_equal(expected[:, 1],
                                       p_arr.pierce['depth'], 1)
        np.testing.assert_almost_equal(expected[:, 2],
                                       p_arr.pierce['time'], 1)
        # NB: we do not check pierce['lat'] and pierce['lon'] here, as these
        # are not calculated when geographiclib is not installed. We check
        # that they are not present.
        with pytest.raises(ValueError):
            p_arr.pierce["lat"]
        with pytest.raises(ValueError):
            p_arr.pierce["lon"]

    def test_vs_java_iasp91(self, taup_testdata):
        """
        Tests the traveltime calculation against the output from TauP in the
        file 'java_tauptime_testoutput'.

        Essentially tests all kinds of depths and epicentral distances in the
        model iasp91.

        Test data generated with:

          $ phases="P,p,PcP,PcS,Pdiff,PKIKKIKP,PKIKKIKS,PKIKP,PKiKP,PKIKPPKIKP"
          $ phases="${phases},PKIKS,PKKP,PKKS,PKP,PKPPKP,PP,pP,pPdiff,pPKIKP"
          $ phases="${phases},pPKiKP,pPKP,PS,pS,pSKIKS,pSKS,S,s,ScP,ScS,Sdiff"
          $ phases="${phases},SKIKKIKP,SKIKKIKS,SKIKP,SKiKP,SKIKS,SKIKSSKIKS"
          $ phases="${phases},SKKP,SKKS,SKS,SKSSKS,SP,sP,sPdiff,sPKIKP,sPKiKP"
          $ phases="${phases},sPKP,SS,sS,sSdiff,sSKIKS,sSKS"

          $ for dist in 0 45 90 180 160; do
              for depth in 0 100 1000 2889; do
                ./taup_time -mod iasp91 -ph ${phases} -h ${depth} -deg ${dist}
              done
            done
        """
        m = TauPyModel(model="iasp91")
        filename = taup_testdata['java_tauptime_testoutput']

        expected = collections.defaultdict(list)
        all_phases = []

        with open(filename, "rt") as fh:
            for line in fh:
                line = line.strip()
                if not line or line.startswith("----"):
                    continue
                line = line.replace("=", "")
                line = line.split()
                try:
                    float(line[0])
                except Exception:
                    continue
                expected[(float(line[0]), float(line[1]))].append({
                    "distance": float(line[0]),
                    "depth": float(line[1]),
                    "name": line[2],
                    "time": float(line[3]),
                    "ray_param_sec_degree": float(line[4]),
                    "takeoff_angle": float(line[5]),
                    "incident_angle": float(line[6]),
                    "purist_distance": float(line[7]),
                    "purist_name": line[8]})
                all_phases.append(line[2])

        all_phases = sorted(set(all_phases))

        for key, value in expected.items():
            expected[key] = sorted(value,
                                   key=lambda x: (x["time"],
                                                  x["ray_param_sec_degree"],
                                                  x["name"]))
        actual_phases = []
        for dist, depth in expected.keys():
            tt = m.get_travel_times(source_depth_in_km=depth,
                                    distance_in_degree=dist,
                                    phase_list=all_phases)
            tt = sorted(tt, key=lambda x: (
                round(x.time, 2),
                round(x.ray_param_sec_degree, 3),
                x.name))
            expected_arrivals = expected[(dist, depth)]

            for actual_arrival, expected_arrival in zip(tt, expected_arrivals):
                self._assert_arrivals_equal(actual_arrival, expected_arrival)
            for arr in tt:
                actual_phases.append(arr.name)
        actual_phases = sorted(set(actual_phases))
        assert actual_phases == all_phases

    def test_pierce_all_phases(self, taup_testdata):
        """
        Tests pierce points against those calculated in TauP.
        """
        filename = taup_testdata['java_taup_pierce_h10_deg35_ttall']
        expected = collections.defaultdict(list)
        with open(filename, "rt") as fh:
            for line in fh:
                line = line.strip()
                if not line:
                    continue
                if line.startswith(">"):
                    current_phase = line[1:].strip().split()[0]
                    continue
                dist, depth, time = list(map(float, line.split()))
                expected[current_phase].append((dist, depth, time))
        expected_phases = sorted(set(expected.keys()))

        m = TauPyModel(model="iasp91")
        arrivals = m.get_pierce_points(source_depth_in_km=10.0,
                                       distance_in_degree=35.0,
                                       phase_list=["ttall"])

        # Make sure the same stuff is available.
        arrival_phases = sorted(set([_i.name for _i in arrivals]))
        assert expected_phases == arrival_phases

        actual = collections.defaultdict(list)
        for arr in arrivals:
            for p in arr.pierce:
                actual[arr.name].append((
                    round(np.degrees(p['dist']), 2),
                    round(p['depth'], 1),
                    round(p['time'], 1)))

        assert sorted(actual.keys()) == sorted(expected.keys())

        for key in actual.keys():
            actual_values = sorted(actual[key])
            expected_values = sorted(expected[key])
            assert actual_values == expected_values

    def test_pierce_add_depth(self):
        """
        Test pierce points requested at a specific depth.
        """
        model = TauPyModel("iasp91")
        depth = 1000.0
        arrivals = model.get_pierce_points(300.0, 50.0, ["PcP"],
                                           add_depth=[depth])
        pierce = arrivals[0].pierce
        assert pierce[3]['depth'] == depth
        assert abs(pierce[3]['dist'] - 0.051) < 2e-3

    def test_single_path_iasp91(self, taup_testdata):
        """
        Test the raypath for a single phase.
        """
        filename = taup_testdata["taup_path_-o_stdout_-h_10_-ph_P_-deg_35"]
        expected = np.genfromtxt(filename, comments='>')

        m = TauPyModel(model="iasp91")
        arrivals = m.get_ray_paths(source_depth_in_km=10.0,
                                   distance_in_degree=35.0, phase_list=["P"])
        assert len(arrivals) == 1

        # Interpolate both paths to 100 samples and make sure they are
        # approximately equal.
        sample_points = np.linspace(0, 35, 100)

        interpolated_expected = np.interp(
            sample_points,
            expected[:, 0],
            expected[:, 1])

        interpolated_actual = np.interp(
            sample_points,
            np.round(np.degrees(arrivals[0].path['dist']), 2),
            np.round(6371 - arrivals[0].path['depth'], 2))

        assert np.allclose(interpolated_actual,
               interpolated_expected, rtol=1E-4, atol=0)

    @pytest.mark.skipif(
        not geodetics.GEOGRAPHICLIB_VERSION_AT_LEAST_1_34,
        reason='test needs geographiclib >= 1.34')
    def test_single_path_geo_iasp91(self, taup_testdata):
        """
        Test the raypath for a single phase given geographical input.

        This tests the case when geographiclib is installed.
        """
        filename = taup_testdata[
            "taup_path_-mod_iasp91_-o_stdout_-h_10_" +
            "-ph_P_-sta_-45_-60_evt_-80_-60"]
        expected = np.genfromtxt(filename, comments='>')

        m = TauPyModel(model="iasp91")
        with warnings.catch_warnings(record=True):
            warnings.simplefilter("always")
            arrivals = m.get_ray_paths_geo(
                source_depth_in_km=10.0,
                source_latitude_in_deg=-80.0,
                source_longitude_in_deg=-60.0,
                receiver_latitude_in_deg=-45.0,
                receiver_longitude_in_deg=-60.0, phase_list=["P"])
        assert len(arrivals) == 1

        # Interpolate both paths to 100 samples and make sure they are
        # approximately equal.
        sample_points = np.linspace(0, 35, 100)

        interpolated_expected_depth = np.interp(
            sample_points,
            expected[:, 0],
            expected[:, 1])
        interpolated_expected_lat = np.interp(
            sample_points,
            expected[:, 0],
            expected[:, 2])
        interpolated_expected_lon = np.interp(
            sample_points,
            expected[:, 0],
            expected[:, 3])

        interpolated_actual_depth = np.interp(
            sample_points,
            np.round(np.degrees(arrivals[0].path['dist']), 2),
            np.round(6371 - arrivals[0].path['depth'], 2))
        np.testing.assert_allclose(interpolated_actual_depth,
                                   interpolated_expected_depth,
                                   rtol=1E-4, atol=0)

        if geodetics.GEOGRAPHICLIB_VERSION_AT_LEAST_1_34:
            interpolated_actual_lat = np.interp(
                sample_points,
                np.round(np.degrees(arrivals[0].path['dist']), 2),
                np.round(arrivals[0].path['lat'], 2))
            interpolated_actual_lon = np.interp(
                sample_points,
                np.round(np.degrees(arrivals[0].path['dist']), 2),
                np.round(arrivals[0].path['lon'], 2))
            np.testing.assert_allclose(interpolated_actual_lat,
                                       interpolated_expected_lat,
                                       rtol=1E-4, atol=0)
            np.testing.assert_allclose(interpolated_actual_lon,
                                       interpolated_expected_lon,
                                       rtol=1E-4, atol=0)

    def test_single_path_geo_fallback_iasp91(self, taup_testdata):
        """
        Test the raypath for a single phase given geographical input.

        This version of the test checks that things still work when
        geographiclib is not installed.
        """
        has_geographiclib_real = geodetics.HAS_GEOGRAPHICLIB
        geodetics.HAS_GEOGRAPHICLIB = False
        filename = taup_testdata["taup_path_-o_stdout_-h_10_-ph_P_-deg_35"]
        expected = np.genfromtxt(filename, comments='>')

        m = TauPyModel(model="iasp91")
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            arrivals = m.get_ray_paths_geo(source_depth_in_km=10.0,
                                           source_latitude_in_deg=-80.0,
                                           source_longitude_in_deg=-60.0,
                                           receiver_latitude_in_deg=-45.0,
                                           receiver_longitude_in_deg=-60.0,
                                           phase_list=["P"])
            geodetics.HAS_GEOGRAPHICLIB = has_geographiclib_real
            assert issubclass(w[-1].category, UserWarning)

        assert len(arrivals) == 1

        # Interpolate both paths to 100 samples and make sure they are
        # approximately equal.
        sample_points = np.linspace(0, 35, 100)

        interpolated_expected = np.interp(
            sample_points,
            expected[:, 0],
            expected[:, 1])

        interpolated_actual = np.interp(
            sample_points,
            np.round(np.degrees(arrivals[0].path['dist']), 2),
            np.round(6371 - arrivals[0].path['depth'], 2))

        np.testing.assert_allclose(interpolated_actual, interpolated_expected,
                                   rtol=1E-4, atol=0)

        # NB: we do not check path['lat'] and path['lon'] here, as these
        # are not calculated when geographiclib is not installed. We check
        # that they are not present.
        with pytest.raises(ValueError):
            arrivals[0].path["lat"]
        with pytest.raises(ValueError):
            arrivals[0].path["lon"]

    def test_single_path_ak135(self, taup_testdata):
        """
        Test the raypath for a single phase. This time for model AK135.
        """
        filename = taup_testdata[
            "taup_path_-o_stdout_-h_10_-ph_P_-deg_35_-mod_ak135"]
        expected = np.genfromtxt(filename, comments='>')

        m = TauPyModel(model="ak135")
        arrivals = m.get_ray_paths(source_depth_in_km=10.0,
                                   distance_in_degree=35.0, phase_list=["P"])
        assert len(arrivals) == 1

        # Interpolate both paths to 100 samples and make sure they are
        # approximately equal.
        sample_points = np.linspace(0, 35, 100)

        interpolated_expected = np.interp(
            sample_points,
            expected[:, 0],
            expected[:, 1])

        interpolated_actual = np.interp(
            sample_points,
            np.round(np.degrees(arrivals[0].path['dist']), 2),
            np.round(6371 - arrivals[0].path['depth'], 2))

        assert np.allclose(interpolated_actual,
               interpolated_expected, rtol=1E-3, atol=0)

    def _read_ak135_test_files(self, filename):
        """
        Helper function parsing the AK135 test data from the original TauP
        test suite.
        """
        with open(filename, "rb") as fh:
            line = fh.readline()
            line = line.strip().split()
            depths = list(map(float, line))

            data = np.genfromtxt(fh)

        dist = data[:, 0]
        time = data[:, 1:-len(depths)]
        ray_param = data[:, -len(depths):]
        time = time[:, ::2] * 60.0 + time[:, 1::2]

        values = np.empty(np.size(ray_param),
                          dtype=[('depth', np.float64),
                                 ('dist', np.float64),
                                 ('ray_param', np.float64),
                                 ('time', np.float64)])

        values['depth'] = np.tile(depths, len(dist))
        values['dist'] = np.repeat(dist, len(depths))
        values['ray_param'] = ray_param.flat
        values['time'] = time.flat

        return values

    def _compare_against_ak135_tables_kennet(self, filename, phases):
        """
        Helper function to compare against the AK135 traveltime tables of
        Kennet. This test is also done in the Java TauP version.
        """
        values = self._read_ak135_test_files(filename)
        m = TauPyModel(model="ak135")
        start = np.random.randint(SPEEDUP_FACTOR)
        for value in values[start::SPEEDUP_FACTOR]:
            # Parameters are not strictly defined for a non-existent travel
            # time.
            if value["time"] == 0.0:
                continue
            arrivals = m.get_ray_paths(
                source_depth_in_km=value["depth"],
                distance_in_degree=value["dist"],
                phase_list=phases)
            arrivals = sorted(arrivals, key=lambda x: x.time)
            arr = arrivals[0]
            # These are the same tolerances as in the Java tests suite.
            assert abs(arr.time - value["time"]) < 0.07
            assert abs(arr.ray_param_sec_degree - value["ray_param"]) < 0.11

    def test_kennet_ak135_ttime_tables_p_deep(self, taup_testdata):
        self._compare_against_ak135_tables_kennet(
            taup_testdata["ak135_P_deep.txt"], phases=["p", "Pdiff", "P"])

    def test_kennet_ak135_ttime_tables_p_shallow(self, taup_testdata):
        self._compare_against_ak135_tables_kennet(
            taup_testdata["ak135_P_shallow.txt"], phases=["p", "Pdiff", "P"])

    def test_kennet_ak135_ttime_tables_pcp(self, taup_testdata):
        self._compare_against_ak135_tables_kennet(
            taup_testdata["ak135_PcP.txt"], phases=["PcP"])

    def test_kennet_ak135_ttime_tables_pkikp(self, taup_testdata):
        self._compare_against_ak135_tables_kennet(
            taup_testdata["ak135_PKIKP.txt"], phases=["PKIKP"])

    def test_kennet_ak135_ttime_tables_s_deep(self, taup_testdata):
        self._compare_against_ak135_tables_kennet(
            taup_testdata["ak135_S_deep.txt"], phases=["s", "S", "Sdiff"])

    def test_kennet_ak135_ttime_tables_s_shallow(self, taup_testdata):
        self._compare_against_ak135_tables_kennet(
            taup_testdata["ak135_S_shallow.txt"], phases=["s", "S", "Sdiff"])

    def test_kennet_ak135_ttime_tables_scp(self, taup_testdata):
        self._compare_against_ak135_tables_kennet(
            taup_testdata["ak135_ScP.txt"], phases=["ScP"])

    def test_kennet_ak135_ttime_tables_scs(self, taup_testdata):
        self._compare_against_ak135_tables_kennet(
            taup_testdata["ak135_ScS.txt"], phases=["ScS"])

    def test_java_pnps(self, taup_testdata):
        """
        Test for Pn Ps waves bug, 53797854298c6ee9dcbf3398bbec3bdd12def964

        Reference output generated by::

          $ ./taup_time -mod iasp91 -ph P,S,PP,PPP,PPPP,S,SS,SS,SSSS,Sn,Pn \
              -h 15 -deg 10
        """
        m = TauPyModel(model="iasp91")
        ph = "P,S,PP,PPP,PPPP,S,SS,SS,SSSS,Sn,Pn".split(",")
        arrivals = m.get_travel_times(source_depth_in_km=15.0,
                                      distance_in_degree=10.0,
                                      phase_list=ph)
        self._compare_arrivals_with_file(
            arrivals, taup_testdata["java_tauptime_pnsn"])

    def test_surface_wave_ttimes(self, taup_testdata):
        """
        Tests the calculation of surface ttimes.

        Tested against a reference output from the Java TauP version.
        """
        for model, table in [("iasp91", "iasp91_surface_waves_table.txt"),
                             ("ak135", "ak135_surface_waves_table.txt")]:
            m = TauPyModel(model=model)
            filename = taup_testdata[table]

            with open(filename, "rt") as fh:
                for line in fh:
                    _, distance, depth, phase, time, ray_param, _, _ \
                        = line.split()
                    distance, depth, time, ray_param = \
                        map(float, [distance, depth, time, ray_param])

                    arrivals = m.get_travel_times(
                        source_depth_in_km=depth, distance_in_degree=distance,
                        phase_list=[phase])

                    assert len(arrivals) > 0

                    # Potentially multiple arrivals. Get the one closest in
                    # time and closest in ray parameter.
                    arrivals = sorted(
                        arrivals,
                        key=lambda x: (abs(x.time - time),
                                       abs(x.ray_param_sec_degree -
                                           ray_param)))
                    arrival = arrivals[0]
                    assert round(arrival.time, 2) == round(time, 2)
                    param1 = round(arrival.ray_param_sec_degree, 2)
                    param2 = round(ray_param, 2)
                    assert param1 == param2

    def test_underside_reflections(self, taup_testdata):
        """
        Tests the calculation of a couple of underside reflection phases.
        """
        m = TauPyModel(model="iasp91")
        # If an interface that is not in the model is used for a phase name,
        # it should snap to the next closest interface. This is reflected in
        # the purist name of the arrivals.
        arrivals = m.get_travel_times(
            source_depth_in_km=10.0, distance_in_degree=90.0,
            phase_list=["P", "PP", "P^410P", "P^660P", "P^300P", "P^400P",
                        "P^500P", "P^600P"])

        self._compare_arrivals_with_file(
            arrivals, taup_testdata['underside_reflections.txt'])

    def test_buried_receiver(self, taup_testdata):
        """
        Simple test for a buried receiver.
        """
        m = TauPyModel(model="iasp91")
        arrivals = m.get_travel_times(
            source_depth_in_km=10.0, distance_in_degree=90.0,
            receiver_depth_in_km=50,
            phase_list=["P", "PP", "S"])

        self._compare_arrivals_with_file(
            arrivals, taup_testdata['buried_receivers.txt'])

    def test_different_models(self):
        """
        Open all included models and make sure that they can produce
        reasonable travel times.
        """
        models = ["1066a", "1066b", "ak135", "herrin", "iasp91", "prem",
                  "sp6", "jb", "pwdk", "ak135f_no_mud"]
        for model in models:
            m = TauPyModel(model=model)

            # Get a p phase.
            arrivals = m.get_travel_times(
                source_depth_in_km=10.0, distance_in_degree=50.0,
                phase_list=["P"])
            # AK135 travel time.
            expected = 534.4
            assert abs(arrivals[0].time - expected) < 5

            # Get an s phase.
            arrivals = m.get_travel_times(
                source_depth_in_km=10.0, distance_in_degree=50.0,
                phase_list=["S"])
            # AK135 travel time.
            expected = 965.1
            # Some models do produce s-waves but they are very far from the
            # AK135 value.
            assert abs(arrivals[0].time - expected) < 50

    def test_paths_for_crustal_phases(self):
        """
        Tests that Pn and PmP are correctly modelled and not mixed up.

        See #1392.
        """
        model = TauPyModel(model='iasp91')
        paths = model.get_ray_paths(source_depth_in_km=0,
                                    distance_in_degree=1,
                                    phase_list=['Pn', 'PmP'])
        assert len(paths) == 2

        assert paths[0].name == "PmP"
        assert round(abs(paths[0].time-21.273), 3) == 0
        assert paths[1].name == "Pn"
        assert round(abs(paths[1].time-21.273), 3) == 0

        assert round(abs(paths[0].time-21.273), 3) == 0

        # Values of visually checked paths to guard against regressions.
        pmp_path = [
            [0.0, 0.0],
            [8.732364066174294e-07, 0.005402127286288305],
            [0.00020293803558129412, 1.2550644943312363],
            [0.000405124234951687, 2.5047268613761844],
            [0.0008098613580518535, 5.004051595465171],
            [0.0016207981804224497, 10.002701063643144],
            [0.00243369214394241, 15.001350531821117],
            [0.0032485517460744207, 20.0],
            [0.0034500021984838003, 20.9375],
            [0.0036515675727796315, 21.875],
            [0.004055043605178164, 23.75],
            [0.004863380445624696, 27.5],
            [0.006485622554890742, 35.0],
            [0.008803860898528547, 35.018603858353345],
            [0.011122099242166353, 35.0],
            [0.012744341351432398, 27.5],
            [0.01355267819187893, 23.75],
            [0.013956154224277463, 21.875],
            [0.014157719598573294, 20.9375],
            [0.014359170050982674, 20.0],
            [0.015174029653114684, 15.001350531821117],
            [0.015986923616634643, 10.002701063643144],
            [0.01679786043900524, 5.004051595465171],
            [0.017202597562105407, 2.5047268613761844],
            [0.0174047837614758, 1.2550644943312363],
            [0.017606848560650475, 0.005402127286288305],
            [0.017607721797057094, 0.0]]
        pn_path = [
            [0.0, 0.0],
            [8.732421799574388e-07, 0.005402127286288305],
            [0.00020293937754080365, 1.2550644943312363],
            [0.0004051269144571584, 2.5047268613761844],
            [0.0008098667167377564, 5.004051595465171],
            [0.0016208089138889542, 10.002701063643144],
            [0.002433708274208186, 15.001350531821117],
            [0.003248573295310095, 20.0],
            [0.0034500255976490177, 20.9375],
            [0.0036515928239481774, 21.875],
            [0.004055072566590625, 23.75],
            [0.004863416852584004, 27.5],
            [0.0064856739540738425, 35.0],
            [0.0064856739540738425, 35.0],
            [0.010967618565869454, 35.0],
            [0.010967618565869454, 35.0],
            [0.012589875667359293, 27.5],
            [0.013398219953352672, 23.75],
            [0.01380169969599512, 21.875],
            [0.01400326692229428, 20.9375],
            [0.014204719224633202, 20.0],
            [0.015019584245735112, 15.001350531821117],
            [0.015832483606054343, 10.002701063643144],
            [0.01664342580320554, 5.004051595465171],
            [0.017048165605486137, 2.5047268613761844],
            [0.017250353142402492, 1.2550644943312363],
            [0.017452419277763337, 0.005402127286288305],
            [0.017453292519943295, 0.0]]

        # Tolerance set to 2e-3 due to limited accuracy of
        # ray path values above
        atol = 2e-3
        np.testing.assert_allclose([_i[0] for _i in pmp_path],
                                   paths[0].path["dist"],
                                   atol=atol)
        np.testing.assert_allclose([_i[1] for _i in pmp_path],
                                   paths[0].path["depth"],
                                   atol=atol)
        np.testing.assert_allclose([_i[0] for _i in pn_path],
                                   paths[1].path["dist"],
                                   atol=atol)
        np.testing.assert_allclose([_i[1] for _i in pn_path],
                                   paths[1].path["depth"],
                                   atol=atol)

    def test_arrivals_class(self):
        """
        Tests list operations on the Arrivals class.

        See #1518.
        """
        model = TauPyModel(model='iasp91')
        arrivals = model.get_ray_paths(source_depth_in_km=0,
                                       distance_in_degree=1,
                                       phase_list=['Pn', 'PmP'])
        assert len(arrivals) == 2
        # test copy
        assert isinstance(arrivals.copy(), Arrivals)
        # test sum
        assert isinstance(arrivals + arrivals, Arrivals)
        assert isinstance(arrivals + arrivals[0], Arrivals)
        # test multiplying
        assert isinstance(arrivals * 2, Arrivals)
        arrivals *= 3
        assert len(arrivals) == 6
        assert isinstance(arrivals, Arrivals)
        # test slicing
        assert isinstance(arrivals[2:5], Arrivals)
        # test appending
        arrivals.append(arrivals[0])
        assert len(arrivals) == 7
        assert isinstance(arrivals, Arrivals)
        # test assignment
        arrivals[0] = arrivals[-1]
        assert isinstance(arrivals, Arrivals)
        arrivals[2:5] = arrivals[1:4]
        assert isinstance(arrivals, Arrivals)
        # test assignment with wrong type
        with pytest.raises(TypeError):
            arrivals[0] = 10.
        with pytest.raises(TypeError):
            arrivals[2:5] = [0, 1, 2]
        with pytest.raises(TypeError):
            arrivals.append(arrivals)
        # test add and mul with wrong type
        with pytest.raises(TypeError):
            arrivals + [2, ]
        with pytest.raises(TypeError):
            arrivals += [2, ]
        with pytest.raises(TypeError):
            arrivals * [2, ]
        with pytest.raises(TypeError):
            arrivals *= [2, ]

    def test_regional_models(self, testdata):
        """
        Tests small regional models as this used to not work.
        """
        model_names = ["1_layer_model", "2_layer_model", "5_layer_model",
                       "2_layer_no_discontinuity_model"]
        expected_results = [
            [("p", 18.143), ("sP", 22.054), ("s", 31.509), ("PP", 4107.380),
             ("SP", 5619.257), ("PS", 5621.396), ("SS", 7133.265)],
            [("p", 18.143), ("P", 19.202), ("Pn", 19.202), ("P", 19.884),
             ("sP", 22.054), ("pP", 23.023), ("pP", 23.038), ("sP", 25.656),
             ("sP", 25.759), ("s", 31.509), ("S", 33.395), ("Sn", 33.395),
             ("S", 34.533), ("sS", 39.991), ("sS", 40.009), ("PP", 3110.537),
             ("SP", 4267.568), ("PS", 4269.707), ("SS", 5426.732)],
            [("P", 17.358), ("Pn", 17.358), ("P", 17.666), ("p", 17.804),
             ("P", 17.869), ("P", 18.039), ("pP", 21.125), ("pP", 21.164),
             ("sP", 22.640), ("sP", 22.716), ("sP", 22.992), ("sP", 23.766),
             ("sP", 23.918), ("sP", 24.039), ("sP", 24.041), ("S", 30.029),
             ("Sn", 30.029), ("S", 30.563), ("s", 30.800), ("S", 30.913),
             ("S", 31.208), ("sS", 36.547), ("sS", 36.613), ("PP", 3147.085),
             ("SP", 4294.699), ("PS", 4296.826), ("SS", 5444.434)],
            [("p", 18.143), ("sP", 22.054), ("s", 31.509), ("PP", 3562.683),
             ("SP", 4881.292), ("PS", 4883.430), ("SS", 6202.037)]
            ]

        for model_name, expects in zip(model_names, expected_results):
            with TemporaryWorkingDirectory():
                folder = Path().absolute()
                build_taup_model(
                    filename=testdata[model_name + ".tvel"],
                    output_folder=folder, verbose=False)
                model = TauPyModel(os.path.join(folder, model_name + ".npz"))

            arrivals = model.get_ray_paths(source_depth_in_km=18.0,
                                           distance_in_degree=1.0)

            assert len(arrivals) == len(expects)
            for arrival, expect in zip(arrivals, expects):
                assert arrival.name == expect[0]
                assert round(abs(arrival.time-expect[1]), 2) == 0

    def test_fruit(self, testdata):
        """
        Check handling of low velocity zones.

        Test case where bottom of mantle has a negative velocity gradient
        so phase cannot turn.
        """
        model_name = "fruit"
        with TemporaryWorkingDirectory():
            folder = os.path.abspath(os.curdir)
            build_taup_model(
              filename=testdata[model_name + ".nd"],
              output_folder=folder, verbose=False)
            model = TauPyModel(os.path.join(folder, model_name + ".npz"))

            s_mod = model.model.s_mod
            high_slow_range = s_mod.high_slowness_layer_depths_p[0]
            assert s_mod.depth_in_high_slowness(1360, 18, True)
            assert s_mod.depth_in_high_slowness(
                1360,
                high_slow_range.ray_param+0.00000001,
                True)
            assert not s_mod.depth_in_high_slowness(
                1360,
                high_slow_range.ray_param,
                True)
            assert not s_mod.depth_in_high_slowness(
                1360,
                high_slow_range.ray_param-0.00000001,
                True)

            p = SeismicPhase("P", model.model)
            atol = 1e-2
            np.testing.assert_allclose(p.min_distance, 0.0, atol=atol)
            np.testing.assert_allclose(p.max_distance*180.0/np.pi,
                                       72.66, atol=atol)
            assert p.branch_seq == [0, 0]

    def test_high_slowness_crust(self, testdata):
        """
        Check handling of sources located in a high slowness layer.
        """
        model_name = "high_slowness_crust"
        with TemporaryWorkingDirectory():
            folder = os.path.abspath(os.curdir)
            build_taup_model(
              filename=testdata[model_name + ".nd"],
              output_folder=folder, verbose=False)
            model = TauPyModel(os.path.join(folder, model_name + ".npz"))
            arrivals = model.get_ray_paths(20.0, 0.84, phase_list=["sS"])
            assert len(arrivals) == 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
The psd test suite.
"""
import gzip
import io
import re
import warnings
from copy import deepcopy

import numpy as np
import pytest

from obspy import Stream, Trace, UTCDateTime, read, read_inventory, Inventory
from obspy.core import Stats
from obspy.core.inventory import Response
from obspy.core.util import CatchAndAssertWarnings
from obspy.core.util.base import NamedTemporaryFile
from obspy.core.util.obspy_types import ObsPyException
from obspy.io.xseed import Parser
from obspy.signal.spectral_estimation import (PPSD, welch_taper, welch_window)
from obspy.signal.spectral_estimation import earthquake_models
from obspy.signal.spectral_estimation import get_idc_infra_low_noise
from obspy.signal.spectral_estimation import get_idc_infra_hi_noise


@pytest.fixture(scope="module")
def _internal_sample_data(root):
    """
    Returns some real data (trace and poles and zeroes) for PPSD testing.

    Data was downsampled to 100Hz so the PPSD is a bit distorted which does
    not matter for the purpose of testing.
    """
    # load test file
    # we reuse this fixture in imaging, so we can't use "testdata" fixture to
    # lookup the path
    file_data = (root / "signal" / "tests" / "data" /
                 'BW.KW1._.EHZ.D.2011.090_downsampled.asc.gz')
    # parameters for the test
    with gzip.open(file_data) as f:
        data = np.loadtxt(f)
    stats = {'_format': 'MSEED',
             'calib': 1.0,
             'channel': 'EHZ',
             'delta': 0.01,
             'endtime': UTCDateTime(2011, 3, 31, 2, 36, 0, 180000),
             'location': '',
             'mseed': {'dataquality': 'D', 'record_length': 512,
                       'encoding': 'STEIM2', 'byteorder': '>'},
             'network': 'BW',
             'npts': 936001,
             'sampling_rate': 100.0,
             'starttime': UTCDateTime(2011, 3, 31, 0, 0, 0, 180000),
             'station': 'KW1'}
    tr = Trace(data, stats)

    paz = {'gain': 60077000.0,
           'poles': [(-0.037004 + 0.037016j), (-0.037004 - 0.037016j),
                     (-251.33 + 0j), (-131.04 - 467.29j),
                     (-131.04 + 467.29j)],
           'sensitivity': 2516778400.0,
           'zeros': [0j, 0j]}

    return tr, paz


@pytest.fixture(scope="function")
def _sample_data(_internal_sample_data):
    tr, paz = _internal_sample_data
    return tr.copy(), deepcopy(paz)


@pytest.fixture(scope="function")
def _ppsd(_sample_data):
    """
    Returns ready computed ppsd for testing purposes.
    """
    tr, paz = _sample_data
    st = Stream([tr])
    ppsd = PPSD(tr.stats, paz, db_bins=(-200, -50, 0.5))
    ppsd.add(st)
    ppsd.calculate_histogram()
    return ppsd


@pytest.mark.usefixtures('ignore_numpy_errors')
class TestPsd:
    """
    Test cases for psd.
    """
    def test_obspy_psd_vs_pitsa(self, testdata):
        """
        Test to compare results of PITSA's psd routine to the
        :func:`matplotlib.mlab.psd` routine wrapped in
        :func:`obspy.signal.spectral_estimation.psd`.
        The test works on 8192 samples long Gaussian noise with a standard
        deviation of 0.1 generated with PITSA, sampling rate for processing in
        PITSA was 100.0 Hz, length of nfft 512 samples. The overlap in PITSA
        cannot be controlled directly, instead only the number of overlapping
        segments can be specified.  Therefore the test works with zero overlap
        to have full control over the data segments used in the psd.
        It seems that PITSA has one frequency entry more, i.e. the psd is one
        point longer. I dont know were this can come from, for now this last
        sample in the psd is ignored.
        """
        from matplotlib.mlab import psd
        sampling_rate = 100.0
        nfft = 512
        noverlap = 0
        file_noise = testdata['pitsa_noise.npy']
        file_psd_pitsa = testdata[
            "pitsa_noise_psd_samprate_100_nfft_512_noverlap_0.npy"]
        noise = np.load(file_noise, allow_pickle=True)
        # in principle to mimic PITSA's results detrend should be specified as
        # some linear detrending (e.g. from matplotlib.mlab.detrend_linear)
        psd_obspy, _ = psd(noise, NFFT=nfft, Fs=sampling_rate,
                           window=welch_taper, noverlap=noverlap,
                           sides="onesided", scale_by_freq=True)

        psd_pitsa = np.load(file_psd_pitsa)

        # mlab's psd routine returns Nyquist frequency as last entry, PITSA
        # seems to omit it and returns a psd one frequency sample shorter.
        psd_obspy = psd_obspy[:-1]

        # test results. first couple of frequencies match not as exactly as all
        # the rest, test them separately with a little more allowance..
        np.testing.assert_array_almost_equal(psd_obspy[:3], psd_pitsa[:3],
                                             decimal=4)
        np.testing.assert_array_almost_equal(psd_obspy[1:5], psd_pitsa[1:5],
                                             decimal=5)
        np.testing.assert_array_almost_equal(psd_obspy[5:], psd_pitsa[5:],
                                             decimal=6)

    def test_welch_window_vs_pitsa(self, testdata):
        """
        Test that the helper function to generate the welch window delivers the
        same results as PITSA's routine.
        Testing both even and odd values for length of window.
        Not testing strange cases like length <5, though.
        """
        welch_even = testdata['pitsa_welch_window_512.npy']
        welch_odd = testdata['pitsa_welch_window_513.npy']

        for file, n in zip((welch_even, welch_odd), (512, 513)):
            window_pitsa = np.load(file)
            window_obspy = welch_window(n)
            np.testing.assert_array_almost_equal(window_pitsa, window_obspy)

    def test_ppsd(self, testdata, _sample_data, _ppsd):
        """
        Test PPSD routine with some real data.
        """
        # paths of the expected result data
        file_histogram = testdata[
            'BW.KW1._.EHZ.D.2011.090_downsampled__ppsd_hist_stack.npy']
        file_binning = testdata[
            'BW.KW1._.EHZ.D.2011.090_downsampled__ppsd_mixed.npz']
        file_mode_mean = testdata[
            'BW.KW1._.EHZ.D.2011.090_downsampled__ppsd_mode_mean.npz']
        tr, _paz = _sample_data
        st = Stream([tr])
        ppsd = _ppsd
        # read results and compare
        result_hist = np.load(file_histogram)
        assert len(ppsd.times_processed) == 4
        assert ppsd.nfft == 65536
        assert ppsd.nlap == 49152
        np.testing.assert_array_equal(ppsd.current_histogram, result_hist)
        # add the same data a second time (which should do nothing at all) and
        # test again - but it will raise UserWarnings, which we omit for now
        with warnings.catch_warnings(record=True):
            warnings.simplefilter('ignore', UserWarning)
            ppsd.add(st)
            np.testing.assert_array_equal(ppsd.current_histogram, result_hist)
        # test the binning arrays
        binning = np.load(file_binning)
        np.testing.assert_array_equal(ppsd.db_bin_edges, binning['spec_bins'])
        np.testing.assert_array_equal(ppsd.period_bin_centers,
                                      binning['period_bins'])

        # test the mode/mean getter functions
        per_mode, mode = ppsd.get_mode()
        per_mean, mean = ppsd.get_mean()
        result_mode_mean = np.load(file_mode_mean)
        np.testing.assert_array_equal(per_mode, result_mode_mean['per_mode'])
        np.testing.assert_array_equal(mode, result_mode_mean['mode'])
        np.testing.assert_array_equal(per_mean, result_mode_mean['per_mean'])
        np.testing.assert_array_equal(mean, result_mode_mean['mean'])

        # test saving and loading of the PPSD (using a temporary file)
        with NamedTemporaryFile(suffix=".npz") as tf:
            filename = tf.name
            # test saving and loading to npz
            ppsd.save_npz(filename)
            ppsd_loaded = PPSD.load_npz(filename)
            ppsd_loaded.calculate_histogram()
            assert len(ppsd_loaded.times_processed) == 4
            assert ppsd_loaded.nfft == 65536
            assert ppsd_loaded.nlap == 49152
            np.testing.assert_array_equal(ppsd_loaded.current_histogram,
                                          result_hist)
            np.testing.assert_array_equal(ppsd_loaded.db_bin_edges,
                                          binning['spec_bins'])
            np.testing.assert_array_equal(ppsd_loaded.period_bin_centers,
                                          binning['period_bins'])

    def test_ppsd_warnings(self, _ppsd, _sample_data):
        """
        Test some warning messages shown by PPSD routine
        """
        # fixture to get data can not refresh on a lower scope than "function",
        # so need to copy ourselves for reusing multiple times inside a
        # function
        def get_trace():
            return deepcopy(_sample_data[0])

        ppsd = _ppsd
        # test warning message if SEED ID is mismatched
        for key in ('network', 'station', 'location', 'channel'):
            tr = get_trace()
            # change starttime, data could then be added if ID and sampling
            # rate match
            tr.stats.starttime += 24 * 3600
            tr.stats[key] = 'XX'
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter('always', UserWarning)
                assert not ppsd.add(tr)
            assert len(w) == 1
            assert str(w[0].message) == \
                'No traces with matching SEED ID in provided stream object.'
        # test warning message if sampling rate is mismatched
        tr = get_trace()
        # change starttime, data could then be added if ID and sampling
        # rate match
        tr.stats.starttime += 24 * 3600
        tr.stats.sampling_rate = 123
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always', UserWarning)
            assert not ppsd.add(tr)
        expected = ('No traces with matching sampling rate in provided '
                    'stream object.')
        assert len(w) == 1
        assert str(w[0].message) == expected

    def test_ppsd_w_iris(self, testdata):
        # Bands to be used this is the upper and lower frequency band pairs
        fres = zip([0.1, 0.05], [0.2, 0.1])

        file_data_anmo = testdata['IUANMO.seed']
        # Read in ANMO data for one day
        st = read(file_data_anmo)

        # Use a canned ANMO response which will stay static
        paz = {'gain': 86298.5, 'zeros': [0, 0],
               'poles': [-59.4313, -22.7121 + 27.1065j, -22.7121 + 27.1065j,
                         -0.0048004, -0.073199],
               'sensitivity': 3.3554 * 10 ** 9}

        # Make an empty PPSD and add the data
        # use highest frequency given by IRIS Mustang noise-pdf web service
        # (0.475683 Hz == 2.10224036 s) as center of first bin, so that we
        # end up with the same bins.
        ppsd = PPSD(st[0].stats, paz, period_limits=(2.10224036, 1400))
        ppsd.add(st)
        ppsd.calculate_histogram()

        # Get the 50th percentile from the PPSD
        (per, perval) = ppsd.get_percentile(percentile=50)
        perinv = 1 / per

        # Read in the results obtained from a Mustang flat file
        file_data_iris = testdata['IRISpdfExample']
        data = np.genfromtxt(
            file_data_iris, comments='#', delimiter=',',
            dtype=[("freq", np.float64),
                   ("power", np.int32),
                   ("hits", np.int32)])
        freq = data["freq"]
        power = data["power"]
        hits = data["hits"]
        # cut data to same period range as in the ppsd we computed
        # (Mustang returns more long periods, probably due to some zero padding
        # or longer nfft in psd)
        num_periods = len(ppsd.period_bin_centers)
        freqdistinct = np.array(sorted(set(freq), reverse=True)[:num_periods])
        # just make sure that we compare the same periods in the following
        # (as we access both frequency arrays by indices from now on)
        np.testing.assert_allclose(freqdistinct, 1 / ppsd.period_bin_centers,
                                   rtol=1e-4)

        # For each frequency pair we want to compare the mean of the bands
        for fre in fres:
            # determine which bins we want to compare
            mask = (fre[0] < perinv) & (perinv < fre[1])

            # Get the values for the bands from the PPSD
            per_val_good_obspy = perval[mask]

            percenlist = []
            # Now we sort out all of the data from the IRIS flat file
            # We will loop through the frequency values and compute a
            # 50th percentile
            for curfreq in freqdistinct[mask]:
                mask_ = curfreq == freq
                tempvalslist = np.repeat(power[mask_], hits[mask_])
                percenlist.append(np.percentile(tempvalslist, 50))
            # Here is the actual test
            np.testing.assert_allclose(np.mean(per_val_good_obspy),
                                       np.mean(percenlist), rtol=0.0, atol=1.2)

    def test_ppsd_w_iris_against_obspy_results(self, testdata):
        """
        Test against results obtained after merging of #1108.
        """
        # Read in ANMO data for one day
        st = read(testdata['IUANMO.seed'])

        # Read in metadata in various different formats
        paz = {'gain': 86298.5, 'zeros': [0, 0],
               'poles': [-59.4313, -22.7121 + 27.1065j, -22.7121 + 27.1065j,
                         -0.0048004, -0.073199],
               'sensitivity': 3.3554 * 10 ** 9}
        resp = testdata['IUANMO.resp']
        parser = Parser(testdata['IUANMO.dataless'])
        inv = read_inventory(testdata['IUANMO.xml'])

        # load expected results, for both only PAZ and full response
        # we need to catch numpy 2.4.0 warnings (#3668)
        with warnings.catch_warnings(record=True):
            warnings.simplefilter('always')
            filename_paz = testdata['IUANMO_ppsd_paz.npz']
            results_paz = PPSD.load_npz(filename_paz, metadata=None,
                                        allow_pickle=True)
            filename_full = testdata['IUANMO_ppsd_fullresponse.npz']
            results_full = PPSD.load_npz(filename_full, metadata=None,
                                         allow_pickle=True)

        # Calculate the PPSDs and test against expected results
        # first: only PAZ
        ppsd = PPSD(st[0].stats, paz)
        ppsd.add(st)
        # commented code to generate the test data:
        # ## np.savez(filename_paz,
        # ##          **dict([(k, getattr(ppsd, k))
        # ##                  for k in PPSD.NPZ_STORE_KEYS]))
        for key in PPSD.NPZ_STORE_KEYS_ARRAY_TYPES:
            np.testing.assert_allclose(
                getattr(ppsd, key), getattr(results_paz, key), rtol=1e-5)
        for key in PPSD.NPZ_STORE_KEYS_LIST_TYPES:
            for got, expected in zip(getattr(ppsd, key),
                                     getattr(results_paz, key)):
                np.testing.assert_allclose(got, expected, rtol=1e-5)
        for key in PPSD.NPZ_STORE_KEYS_SIMPLE_TYPES:
            if key in ["obspy_version", "numpy_version", "matplotlib_version"]:
                continue
            assert getattr(ppsd, key) == getattr(results_paz, key)
        # second: various methods for full response
        for metadata in [parser, inv, resp]:
            ppsd = PPSD(st[0].stats, metadata)
            ppsd.add(st)
            # commented code to generate the test data:
            # ## np.savez(filename_full,
            # ##          **dict([(k, getattr(ppsd, k))
            # ##                  for k in PPSD.NPZ_STORE_KEYS]))
            for key in PPSD.NPZ_STORE_KEYS_ARRAY_TYPES:
                np.testing.assert_allclose(
                    getattr(ppsd, key), getattr(results_full, key), rtol=1e-5)
            for key in PPSD.NPZ_STORE_KEYS_LIST_TYPES:
                for got, expected in zip(getattr(ppsd, key),
                                         getattr(results_full, key)):
                    np.testing.assert_allclose(got, expected, rtol=1e-5)
            for key in PPSD.NPZ_STORE_KEYS_SIMPLE_TYPES:
                if key in ["obspy_version", "numpy_version",
                           "matplotlib_version"]:
                    continue
                assert getattr(ppsd, key) == getattr(results_full, key)

    def test_ppsd_save_and_load_npz(self, _sample_data, _ppsd):
        """
        Test PPSD.load_npz() and PPSD.save_npz()
        """
        _, paz = _sample_data
        ppsd = _ppsd

        # save results to npz file
        with NamedTemporaryFile(suffix=".npz") as tf:
            filename = tf.name
            # test saving and loading an uncompressed file
            ppsd.save_npz(filename)
            ppsd_loaded = PPSD.load_npz(filename, metadata=paz)

        for key in PPSD.NPZ_STORE_KEYS:
            if isinstance(getattr(ppsd, key), np.ndarray) or \
                    key == '_binned_psds':
                np.testing.assert_equal(getattr(ppsd, key),
                                        getattr(ppsd_loaded, key))
            else:
                assert getattr(ppsd, key) == getattr(ppsd_loaded, key)

    def test_ppsd_restricted_stacks(self, testdata, image_path):
        """
        Test PPSD.calculate_histogram() with restrictions to what data should
        be stacked. Also includes image tests.
        """
        # set up a bogus PPSD, with fixed random psds but with real start times
        # of psd pieces, to facilitate testing the stack selection.
        ppsd = PPSD(stats=Stats(dict(sampling_rate=150)), metadata=None,
                    db_bins=(-200, -50, 20.), period_step_octaves=1.4)
        # change data to nowadays used nanoseconds POSIX timestamp
        ppsd._times_processed = [
            UTCDateTime(t)._ns for t in np.load(
                testdata['ppsd_times_processed.npy']).tolist()]
        np.random.seed(1234)
        ppsd._binned_psds = [
            arr for arr in np.random.uniform(
                -200, -50,
                (len(ppsd._times_processed), len(ppsd.period_bin_centers)))]

        # Test callback function that selects a fixed random set of the
        # timestamps.  Also checks that we get passed the type we expect,
        # which is 1D numpy ndarray of int type.
        def callback(t_array):
            assert isinstance(t_array, np.ndarray)
            assert t_array.shape == (len(ppsd._times_processed),)
            assert np.issubdtype(t_array.dtype, np.integer)
            np.random.seed(1234)
            res = np.random.randint(0, 2, len(t_array)).astype(bool)
            return res

        # test several different sets of stack criteria, should cover
        # everything, even with lots of combined criteria
        stack_criteria_list = [
            dict(starttime=UTCDateTime(2015, 3, 8), month=[2, 3, 5, 7, 8]),
            dict(endtime=UTCDateTime(2015, 6, 7), year=[2015],
                 time_of_weekday=[(1, 0, 24), (2, 0, 24), (-1, 0, 11)]),
            dict(year=[2013, 2014, 2016, 2017], month=[2, 3, 4]),
            dict(month=[1, 2, 5, 6, 8], year=2015),
            dict(isoweek=[4, 5, 6, 13, 22, 23, 24, 44, 45]),
            dict(time_of_weekday=[(5, 22, 24), (6, 0, 2), (6, 22, 24)]),
            dict(callback=callback, month=[1, 3, 5, 7]),
            dict(callback=callback)]
        expected_selections = np.load(
            testdata['ppsd_stack_selections.npy'])

        # test every set of criteria
        for stack_criteria, expected_selection in zip(
                stack_criteria_list, expected_selections):
            selection_got = ppsd._stack_selection(**stack_criteria)
            np.testing.assert_array_equal(selection_got, expected_selection)

        plot_kwargs = dict(max_percentage=15, xaxis_frequency=True,
                           period_lim=(0.01, 50))
        ppsd.calculate_histogram(**stack_criteria_list[1])
        fig = ppsd.plot(show=False, **plot_kwargs)

        fig.axes[1].set_xlim(left=fig.axes[1].get_xlim()[0] - 2)
        image_path_1 = image_path.parent / 'test_ppsd_restricted_stacks_1.png'
        with np.errstate(under='ignore'):
            fig.savefig(image_path_1)

        # test it again, checking that updating an existing plot with different
        # stack selection works..
        #  a) we start with the stack for the expected image and test that it
        #     matches (like above):
        ppsd.calculate_histogram(**stack_criteria_list[1])
        image_path_2 = image_path.parent / 'test_ppsd_restricted_stacks_2.png'
        with np.errstate(under='ignore'):
            fig.savefig(image_path_2)

        ppsd.calculate_histogram(**stack_criteria_list[1])
        image_path_3 = image_path.parent / 'test_ppsd_restricted_stacks_3.png'
        ppsd._plot_histogram(fig=fig, draw=True)
        with np.errstate(under='ignore'):
            fig.savefig(image_path_3)

    def test_earthquake_models(self, _ppsd):
        """
        Test earthquake models
        """
        ppsd = _ppsd
        test_magnitudes = [3.5, 2.5, 1.5]
        distance = 10
        for magnitude in test_magnitudes:
            key = (magnitude, distance)
            fig = ppsd.plot(
                show_earthquakes=(magnitude - 0.5, magnitude + 0.5, 5, 15),
                show_noise_models=False, show=False)
            ax = fig.axes[0]
            line = ax.lines[0]
            frequencies, accelerations = earthquake_models[key]
            accelerations = np.array(accelerations)
            periods = 1 / np.array(frequencies)
            power = accelerations / (periods ** (-.5))
            power = 20 * np.log10(power / 2)
            assert list(line.get_ydata()) == list(power)
            assert list(line.get_xdata()) == list(periods)
            caption = 'M%.1f\n%dkm' % (magnitude, distance)
            assert ax.texts[0].get_text() == caption

    def test_ppsd_infrasound(self, testdata):
        """
        Test plotting psds on infrasound data
        """
        wf = testdata['IM.I59H1..BDF_2020_10_31.mseed']
        md = testdata['IM.I59H1..BDF_2020_10_31.xml']
        st = read(wf)
        inv = read_inventory(md)
        tr = st[0]
        ppsd = PPSD(tr.stats, metadata=inv, special_handling='infrasound',
                    db_bins=(-100, 40, 1.), ppsd_length=300, overlap=0.5)
        ppsd.add(st)
        fig = ppsd.plot(xaxis_frequency=True,  period_lim=(0.01, 10),
                        show=False)
        models = (get_idc_infra_hi_noise(), get_idc_infra_low_noise())
        lines = fig.axes[0].lines
        freq1 = lines[0].get_xdata()
        per1 = 1 / freq1
        hn = lines[0].get_ydata()
        freq2 = lines[1].get_xdata()
        per2 = 1 / freq2
        ln = lines[1].get_ydata()
        per1_m, hn_m = models[0]
        per2_m, ln_m = models[1]
        np.testing.assert_array_equal(hn, hn_m)
        np.testing.assert_array_equal(ln, ln_m)
        np.testing.assert_array_equal(per1, per1_m)
        np.testing.assert_array_equal(per2, per2_m)
        # test calculated psd values
        psd = ppsd.psd_values[0]
        assert len(psd) == 73
        psd = psd[:20]
        expected = np.array([
            -63.424206, -64.07918, -64.47593, -64.77374, -65.09937,
            -67.17343, -66.36576, -65.75002, -65.34155, -64.58012,
            -63.72327, -62.615784, -61.612656, -61.085754, -60.09534,
            -58.949272, -57.600315, -56.43776, -55.448067, -54.242218],
            dtype=np.float32)
        np.testing.assert_array_almost_equal(psd, expected, decimal=2)

    def test_ppsd_add_npz(self, testdata):
        """
        Test PPSD.add_npz().
        """
        # set up a bogus PPSD, with fixed random psds but with real start times
        # of psd pieces, to facilitate testing the stack selection.
        ppsd = PPSD(stats=Stats(dict(sampling_rate=150)), metadata=None,
                    db_bins=(-200, -50, 20.), period_step_octaves=1.4)
        _times_processed = np.load(
            testdata['ppsd_times_processed.npy']).tolist()
        # change data to nowadays used nanoseconds POSIX timestamp
        _times_processed = [UTCDateTime(t)._ns for t in _times_processed]
        np.random.seed(1234)
        _binned_psds = [
            arr for arr in np.random.uniform(
                -200, -50,
                (len(_times_processed), len(ppsd.period_bin_centers)))]

        with NamedTemporaryFile(suffix=".npz") as tf1, \
                NamedTemporaryFile(suffix=".npz") as tf2, \
                NamedTemporaryFile(suffix=".npz") as tf3:
            # save data split up over three separate temporary files
            ppsd._times_processed = _times_processed[:200]
            ppsd._binned_psds = _binned_psds[:200]
            ppsd.save_npz(tf1.name)
            ppsd._times_processed = _times_processed[200:400]
            ppsd._binned_psds = _binned_psds[200:400]
            ppsd.save_npz(tf2.name)
            ppsd._times_processed = _times_processed[400:]
            ppsd._binned_psds = _binned_psds[400:]
            ppsd.matplotlib_version = "X.X.X"
            ppsd.save_npz(tf3.name)
            # now load these saved npz files and check if all data is present
            ppsd = PPSD.load_npz(tf1.name, metadata=None)
            ppsd.add_npz(tf2.name)
            # we changed a version number so this should emit a warning
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter('always')
                ppsd.add_npz(tf3.name)
                assert len(w) == 1
            np.testing.assert_array_equal(_binned_psds, ppsd._binned_psds)
            np.testing.assert_array_equal(_times_processed,
                                          ppsd._times_processed)
            # adding data already present should also emit a warning and the
            # PPSD should not be changed
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter('always')
                ppsd.add_npz(tf2.name)
                assert len(w) == 1
            np.testing.assert_array_equal(_binned_psds, ppsd._binned_psds)
            np.testing.assert_array_equal(_times_processed,
                                          ppsd._times_processed)

    def test_ppsd_time_checks(self):
        """
        Some tests that make sure checking if a new PSD slice to be addded to
        existing PPSD has an invalid overlap or not works as expected.
        """
        # Calling PPSD with an empty Response object will emit a warning
        # that it's impossible to compute evalresp (in _preload_responses)
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')
            ppsd = PPSD(Stats(), Response())
            assert len(w) == 1
        one_second = 1000000000
        t0 = 946684800000000000  # 2000-01-01T00:00:00
        time_diffs = [
            0, one_second, one_second * 2, one_second * 3,
            one_second * 8, one_second * 9, one_second * 10]
        ppsd._times_processed = [t0 + td for td in time_diffs]
        ppsd.ppsd_length = 2
        ppsd.overlap = 0.5
        # valid time stamps to insert data for (i.e. data that overlaps with
        # existing data at most "overlap" times "ppsd_length")
        ns_ok = [
            t0 - 3 * one_second,
            t0 - 1.01 * one_second,
            t0 - one_second,
            t0 + 4 * one_second,
            t0 + 4.01 * one_second,
            t0 + 6 * one_second,
            t0 + 7 * one_second,
            t0 + 6.99 * one_second,
            t0 + 11 * one_second,
            t0 + 11.01 * one_second,
            t0 + 15 * one_second,
            ]
        for ns in ns_ok:
            t = UTCDateTime(ns=int(ns))
            # getting False means time is not present yet and a PSD slice would
            # be added to the PPSD data
            assert not ppsd._PPSD__check_time_present(t)
        # invalid time stamps to insert data for (i.e. data that overlaps with
        # existing data more than "overlap" times "ppsd_length")
        ns_bad = [
            t0 - 0.99 * one_second,
            t0 - 0.5 * one_second,
            t0,
            t0 + 1.1 * one_second,
            t0 + 3.99 * one_second,
            t0 + 7.01 * one_second,
            t0 + 7.5 * one_second,
            t0 + 8 * one_second,
            t0 + 8.8 * one_second,
            t0 + 10 * one_second,
            t0 + 10.99 * one_second,
            ]
        for ns in ns_bad:
            t = UTCDateTime(ns=int(ns))
            # getting False means time is not present yet and a PSD slice would
            # be added to the PPSD data
            assert ppsd._PPSD__check_time_present(t)

    def test_issue1216(self, _sample_data):
        tr, paz = _sample_data
        st = Stream([tr])
        ppsd = PPSD(tr.stats, paz, db_bins=(-200, -50, 0.5))
        ppsd.add(st)
        # After adding data internal representation of hist stack is None
        assert ppsd._current_hist_stack is None
        # Accessing the current_histogram property calculates the default stack
        assert ppsd.current_histogram is not None
        assert ppsd._current_hist_stack is not None
        # Adding the same data again does not invalidate the internal stack
        # but raises "UserWarning: Already covered time spans detected"
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always', UserWarning)
            ppsd.add(st)
            msg = 'Already covered time spans detected'
            w2 = [w_ for w_ in w if str(w_.message).startswith(msg)]
            assert len(w2) == 4
        assert ppsd._current_hist_stack is not None
        # Adding new data invalidates the internal stack
        tr.stats.starttime += 3600
        st2 = Stream([tr])
        # raises "UserWarning: Already covered time spans detected"
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always', UserWarning)
            ppsd.add(st2)
            msg = 'Already covered time spans detected'
            w2 = [w_ for w_ in w if str(w_.message).startswith(msg)]
            assert len(w2) == 2
        assert ppsd._current_hist_stack is None
        # Accessing current_histogram again calculates the stack
        assert ppsd.current_histogram is not None
        assert ppsd._current_hist_stack is not None

    def test_wrong_trace_id_message(self, _sample_data, testdata):
        """
        Test that we get the expected warning message on waveform/metadata
        mismatch.
        """
        tr, _paz = _sample_data
        inv = read_inventory(testdata['IUANMO.xml'])
        st = Stream([tr])
        ppsd = PPSD(tr.stats, inv)
        # metadata doesn't fit the trace ID specified via stats
        # should show a warning..
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')
            ret = ppsd.add(st)
            # the trace is sliced into four segments, so we get the warning
            # message four times..
            assert len(w) == 4
            for w_ in w:
                assert str(w_.message).startswith(
                    "Error getting response from provided metadata")
        # should not add the data to the ppsd
        assert not ret

    def test_ppsd_psd_values(self, _ppsd):
        """
        Test property psd values
        """
        ppsd = _ppsd
        # just test against existing low level data access
        assert ppsd.psd_values == ppsd._binned_psds
        np.testing.assert_array_equal(ppsd.psd_values, ppsd._binned_psds)
        # property can't be set
        with pytest.raises(AttributeError):
            ppsd.psd_values = 123

    def test_ppsd_temporal_plot(self, testdata, image_path):
        """
        Test plot of several period bins over time
        """
        # we need to catch numpy 2.4.0 warnings (#3668)
        with warnings.catch_warnings(record=True):
            warnings.simplefilter('always')
            ppsd = PPSD.load_npz(testdata['ppsd_kw1_ehz.npz'],
                                 allow_pickle=True)

        restrictions = {'starttime': UTCDateTime(2011, 2, 6, 1, 1),
                        'endtime': UTCDateTime(2011, 2, 7, 21, 12),
                        'year': [2011],
                        'time_of_weekday': [(-1, 2, 23)]}

        # add some gaps in the middle
        for i in sorted(list(range(30, 40)) + list(range(8, 18)) + [4])[::-1]:
            ppsd._times_processed.pop(i)
            ppsd._binned_psds.pop(i)
        fig = ppsd.plot_temporal([0.1, 1, 10], filename=None, show=False,
                                 **restrictions)
        fig.savefig(image_path)

    def test_exclude_last_sample(self):
        start = UTCDateTime("2017-01-01T00:00:00")
        header = {
            "starttime": start,
            "network": "GR",
            "station": "FUR",
            "channel": "BHZ"
        }
        # 49 segments of 30 minutes to allow 30 minutes overlap in next day
        tr = Trace(data=np.arange(30 * 60 * 4, dtype=np.int32), header=header)

        ppsd = PPSD(tr.stats, read_inventory())
        ppsd.add(tr)

        assert 3 == len(ppsd._times_processed)
        assert 3600 == ppsd.len
        for i, time in enumerate(ppsd._times_processed):
            current = start.ns + (i * 30 * 60) * 1e9
            assert time == current

    def test_ppsd_spectrogram_plot(self, testdata, image_path):
        """
        Test spectrogram type plot of PPSD
        """
        # we need to catch numpy 2.4.0 warnings (#3668)
        with warnings.catch_warnings(record=True):
            warnings.simplefilter('always')
            ppsd = PPSD.load_npz(testdata['ppsd_kw1_ehz.npz'],
                                 allow_pickle=True)

        # add some gaps in the middle
        for i in sorted(list(range(30, 40)) + list(range(8, 18)) + [4])[::-1]:
            ppsd._times_processed.pop(i)
            ppsd._binned_psds.pop(i)

        ppsd.plot_spectrogram(filename=image_path, show=False)

    def test_ppsd_uneven_sampling_rate(self):
        """
        Regression test for #3387

        Makes sure that with weird sampling rates that do not align with the
        length of the PPSD slice at all, still no data slices are left out.
        The fix for this was to make sure to insert the times used for slicing
        into the processed times list and not the actual sample start time of
        the trace that resulted from slicing based on these times.
        """
        tr = read()[0]
        inv = read_inventory()
        # make the sampling rate a bit weird so we end up with sample start
        # times creeping on a decimal seconds and not aligning well with the
        # ppsd length well
        tr.stats.sampling_rate = 100.1
        ppsd = PPSD(tr.stats, inv, ppsd_length=8, overlap=0.0)
        ppsd.add(tr)
        assert len(ppsd._times_processed) == 3
        assert ppsd._times_processed == [
            1251073203000000000, 1251073211000000000, 1251073219000000000]
        tr.stats.sampling_rate = 99.99
        ppsd = PPSD(tr.stats, inv, ppsd_length=8, overlap=0.0)
        ppsd.add(tr)
        assert len(ppsd._times_processed) == 3
        assert ppsd._times_processed == [
            1251073203000000000, 1251073211000000000, 1251073219000000000]

    def test_exception_reading_newer_npz(self, testdata):
        """
        Checks that an exception is properly raised when trying to read a npz
        that was written on a more recent ObsPy version (specifically that has
        a higher 'ppsd_version' number which is used to keep track of changes
        in PPSD and the npz file used for serialization).
        """
        msg = ("Trying to read/add a PPSD npz with 'ppsd_version=100'. This "
               "file was written on a more recent ObsPy version that very "
               "likely has incompatible changes in PPSD internal structure "
               "and npz serialization. It can not safely be read with this "
               "ObsPy version (current 'ppsd_version' is {!s}). Please "
               "consider updating your ObsPy installation.".format(
                   PPSD(stats=Stats(), metadata=None).ppsd_version))
        # we need to catch numpy 2.4.0 warnings (#3668)
        with warnings.catch_warnings(record=True):
            warnings.simplefilter('always')
            # 1 - loading a npz
            data = np.load(testdata['ppsd_kw1_ehz.npz'], allow_pickle=True)
        # we have to load, modify 'ppsd_version' and save the npz file for the
        # test..
            # we need to catch numpy 2.4.0 warnings (#3668)
        with warnings.catch_warnings(record=True):
            warnings.simplefilter('always')
            items = {key: data[key] for key in data.files}
        # deliberately set a higher ppsd_version number
        items['ppsd_version'] = items['ppsd_version'].copy()
        items['ppsd_version'].fill(100)
        with NamedTemporaryFile() as tf:
            filename = tf.name
            with open(filename, 'wb') as fh:
                np.savez(fh, **items)
            with pytest.raises(ObsPyException, match=re.escape(msg)):
                PPSD.load_npz(filename)
        # 2 - adding a npz
        # we need to catch numpy 2.4.0 warnings (#3668)
        with warnings.catch_warnings(record=True):
            warnings.simplefilter('always')
            ppsd = PPSD.load_npz(testdata['ppsd_kw1_ehz.npz'],
                                 allow_pickle=True)
        for method in (ppsd.add_npz, ppsd._add_npz):
            with NamedTemporaryFile() as tf:
                filename = tf.name
                with open(filename, 'wb') as fh:
                    np.savez(fh, **items)
                with pytest.raises(ObsPyException, match=re.escape(msg)):
                    method(filename)

    def test_nice_ringlaser_metadata_error_msg(self):
        expected = ("When using `special_handling='ringlaser'`, `metadata` "
                    "must be a plain dictionary with key 'sensitivity' "
                    "stating the overall sensitivity`.")
        with pytest.raises(TypeError, match=re.escape(expected)):
            PPSD(stats=Stats(), metadata=Inventory(networks=[], source=""),
                 special_handling='ringlaser')

    @staticmethod
    def _save_npz_require_pickle(filename, ppsd):
        """ Save npz in such a way that requires pickle to load"""
        out = {}
        for key in PPSD.NPZ_STORE_KEYS:
            out[key] = getattr(ppsd, key)
        np.savez_compressed(filename, **out)

    def test_can_read_npz_without_pickle(self, testdata, _ppsd):
        """
        Ensures that a default PPSD can be written and read without having to
        allow np.load the use of pickle, or that a helpful error message is
        raised if allow_pickle is required. See #2409.
        """
        # we need to catch numpy 2.4.0 warnings (#3668)
        with warnings.catch_warnings(record=True):
            warnings.simplefilter('always')
            # Init a test PPSD and empty byte stream.
            ppsd = PPSD.load_npz(testdata['ppsd_kw1_ehz.npz'],
                                 allow_pickle=True)
        byte_me = io.BytesIO()
        # Save PPSD to byte stream and rewind to 0.
        ppsd.save_npz(byte_me)
        byte_me.seek(0)
        # Load dict, will raise an exception if pickle is needed.
        loaded_dict = dict(np.load(byte_me, allow_pickle=False))
        assert isinstance(loaded_dict, dict)
        # A helpful error message is issued when allow_pickle is needed.
        with pytest.raises(ValueError, match='Loading PPSD results'):
            PPSD.load_npz(testdata['ppsd_kw1_ehz.npz'])

        ppsd = _ppsd
        # save PPSD in such a way to mock old versions.
        with NamedTemporaryFile(suffix='.npz') as ntemp:
            temp_path = ntemp.name
            self._save_npz_require_pickle(temp_path, ppsd)
            # We should be able to load the files when allowing pickle.
            PPSD.load_npz(temp_path, allow_pickle=True)
            # If not allow_pickle,  a helpful error msg should be raised.
            with pytest.raises(ValueError, match='Loading PPSD results'):
                PPSD.load_npz(temp_path)

    @pytest.mark.filterwarnings('ignore:.*time ranges already covered.*')
    def test_can_add_npz_without_pickle(self, _ppsd):
        """
        Ensure PPSD can be added without using the pickle protocol, or
        that a helpful error message is raised if allow_pickle is required.
        See #2409.
        """
        ppsd = _ppsd
        # save PPSD in such a way to mock old versions.
        with NamedTemporaryFile(suffix='.npz') as ntemp:
            temp_path = ntemp.name
            self._save_npz_require_pickle(temp_path, ppsd)
            # We should be able to load the files when allowing pickle.
            # we need to catch numpy 2.4.0 warnings (#3668)
            with warnings.catch_warnings(record=True):
                warnings.simplefilter('always')
                ppsd.add_npz(temp_path, allow_pickle=True)
            # If not allow_pickle,  a helpful error msg should be raised.
            with pytest.raises(ValueError, match='Loading PPSD results'):
                ppsd.add_npz(temp_path)

    def test_short_trace_warning(self, _sample_data):
        """
        Makes sure a warning is shown if a trace shorter than ppsd length is
        added which is doing nothing. see #2386
        """
        tr, paz = _sample_data
        ppsd = PPSD(stats=tr.stats, metadata=paz)
        tr.data = tr.data[:4]  # shorten trace
        msg = (f"Trace is shorter than this PPSD's 'ppsd_length' "
               f"({str(ppsd.ppsd_length)} seconds). Skipping trace: "
               f"{str(tr)}")
        with CatchAndAssertWarnings(expected=[(UserWarning, msg)]):
            success = ppsd.add(tr)
        assert not success  # add returns False on noop
        assert not len(ppsd.times_processed)  # should be empty, nothing added
        # contains start/end times of traces added in, even if not processed
        assert len(ppsd.times_data) == 1



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from obspy.scripts.flinnengdahl import main as obspy_flinnengdahl
from obspy.geodetics import FlinnEngdahl
from obspy.core.util.misc import CatchOutput


class TestUtilFlinnEngdahl:

    def test_coordinates(self, testdata):
        flinn_engdahl = FlinnEngdahl()
        with open(testdata['flinnengdahl.csv'], 'r') as fh:
            for line in fh:
                longitude, latitude, checked_region = line.strip().split('\t')
                longitude = float(longitude)
                latitude = float(latitude)

                region = flinn_engdahl.get_region(longitude, latitude)
                assert region == \
                    checked_region, \
                    "(%f, %f) got %s instead of %s" % (
                        longitude,
                        latitude,
                        region,
                        checked_region
                    )

    def test_script(self, testdata):
        with open(testdata['flinnengdahl.csv'], 'r') as fh:
            # Testing once is sufficient.
            line = fh.readline()
            longitude, latitude, checked_region = line.strip().split('\t')

            with CatchOutput() as out:
                obspy_flinnengdahl([longitude, latitude])
            region = out.stdout.strip()

            assert region == \
                checked_region, \
                '(%s, %s) got %s instead of %s' % (
                    longitude,
                    latitude,
                    region,
                    checked_region
                )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
A command-line program that converts XML-SEED into Dataless SEED files.
"""
import os
import sys
from argparse import ArgumentParser
from glob import glob

from obspy import __version__
from obspy.io.xseed.parser import Parser


def xseed2dataless(filename, options):
    files = []
    for item in filename:
        files.extend(glob(item))
    outdir = False
    outfile = False
    if options.output:
        if os.path.isdir(options.output):
            outdir = options.output
        elif len(files) > 1:
            msg = 'More than one filename is given.' + os.linesep
            msg += '\t--output argument will not be used.\n'
            sys.stdout.write(msg)
        else:
            outfile = options.output
    if options.verbose:
        msg = 'Found %s files.' % len(files) + os.linesep
        sys.stdout.write(msg)
    for file in files:
        if not os.path.isfile(file):
            continue
        f = open(file, 'rb')
        if f.read(1) != b'<':
            if options.verbose:
                msg = 'Skipping file %s' % file
                msg += '\t-- not a XML-SEED file' + os.linesep
                sys.stdout.write(msg)
            f.close()
            continue
        f.close()
        if outdir:
            output = os.path.join(outdir,
                                  os.path.basename(file) + os.extsep +
                                  'dataless')
        elif outfile:
            output = outfile
        else:
            output = os.path.basename(file) + os.extsep + 'dataless'
        if options.verbose:
            msg = 'Parsing file %s' % file + os.linesep
            sys.stdout.write(msg)
        try:
            parser = Parser(file, debug=options.debug)
            parser.write_seed(output)
        except Exception as e:
            if options.debug:
                raise
            msg = '\tError parsing file %s' % file + os.linesep
            msg += '\t' + str(e) + os.linesep
            sys.stderr.write(msg)


def main(argv=None):
    parser = ArgumentParser(prog='obspy-xseed2dataless',
                            description=__doc__.strip())
    parser.add_argument('-V', '--version', action='version',
                        version='%(prog)s ' + __version__)
    parser.add_argument('-d', '--debug', action='store_true',
                        help='show debugging information')
    parser.add_argument('-q', '--quiet', action='store_false', dest='verbose',
                        help='non verbose mode')
    parser.add_argument('-o', '--output', default=None,
                        help='output filename or directory')
    parser.add_argument('files', nargs='+', help='files to convert')
    args = parser.parse_args(argv)

    xseed2dataless(args.files, args)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Author: Douglas Creager <dcreager@dcreager.net>
# This file is placed into the public domain.

# Calculates the current version number.  If possible, this is the
# output of “git describe”, modified to conform to the versioning
# scheme that setuptools uses.  If “git describe” returns an error
# (most likely because we're in an unpacked copy of a release tarball,
# rather than in a git working copy), then we fall back on reading the
# contents of the RELEASE-VERSION file.
#
# To use this script, simply import it your setup.py file, and use the
# results of get_git_version() as your package version:
#
# from version import *
#
# setup(
#     version=get_git_version(),
#     .
#     .
#     .
# )
#
# This will automatically update the RELEASE-VERSION file, if
# necessary.  Note that the RELEASE-VERSION file should *not* be
# checked into git; please add it to your top-level .gitignore file.
#
# You'll probably want to distribute the RELEASE-VERSION file in your
# sdist tarballs; to do this, just create a MANIFEST.in file that
# contains the following line:
#
#   include RELEASE-VERSION

# NO IMPORTS FROM OBSPY OR FUTURE IN THIS FILE! (file gets used at
# installation time)
import inspect
import io
import os
import re
from subprocess import STDOUT, CalledProcessError, check_output
import warnings


__all__ = ["get_git_version"]

script_dir = os.path.abspath(os.path.dirname(inspect.getfile(
                                             inspect.currentframe())))
OBSPY_ROOT = os.path.abspath(os.path.join(script_dir, os.pardir,
                                          os.pardir, os.pardir))
VERSION_FILE = os.path.join(OBSPY_ROOT, "obspy", "RELEASE-VERSION")


def call_git_describe(abbrev=10, dirty=True,
                      append_remote_tracking_branch=True):
    try:
        p = check_output(['git', 'rev-parse', '--show-toplevel'],
                         cwd=OBSPY_ROOT, stderr=STDOUT)
        path = p.decode().strip()
    except (OSError, CalledProcessError):
        return None

    if os.path.normpath(path) != OBSPY_ROOT:
        return None

    command = ['git', 'describe', '--abbrev=%d' % abbrev, '--always', '--tags']
    if dirty:
        command.append("--dirty")
    try:
        p = check_output(['git', 'describe', '--dirty', '--abbrev=%d' % abbrev,
                          '--always', '--tags'],
                         cwd=OBSPY_ROOT, stderr=STDOUT)
        line = p.decode().strip()
    except (OSError, CalledProcessError):
        return None

    remote_tracking_branch = None
    if append_remote_tracking_branch:
        try:
            # find out local alias of remote and name of remote tracking branch
            p = check_output(['git', 'branch', '-vv'],
                             cwd=OBSPY_ROOT, stderr=STDOUT)
            remote_info = [line_.rstrip()
                           for line_ in p.decode().splitlines()]
            remote_info = [line_ for line_ in remote_info
                           if line_.startswith('*')][0]
            remote_info = re.sub(r".*? \[([^ :]*).*?\] .*", r"\1", remote_info)
            remote, branch = remote_info.split("/")
            # find out real name of remote
            p = check_output(['git', 'remote', '-v'],
                             cwd=OBSPY_ROOT, stderr=STDOUT)
            stdout = [line_.strip() for line_ in p.decode().splitlines()]
            remote = [line_ for line_ in stdout
                      if line_.startswith(remote)][0].split()[1]
            if remote.startswith("git@github.com:"):
                remote = re.sub(r"git@github.com:(.*?)/.*", r"\1", remote)
            elif remote.startswith("https://github.com/"):
                remote = re.sub(r"https://github.com/(.*?)/.*", r"\1", remote)
            elif remote.startswith("git://github.com"):
                remote = re.sub(r"git://github.com/(.*?)/.*", r"\1", remote)
            else:
                remote = None
            if remote is not None:
                remote_tracking_branch = re.sub(r'[^A-Za-z0-9._-]', r'_',
                                                '%s-%s' % (remote, branch))
        except (IndexError, OSError, ValueError, CalledProcessError):
            pass

    # (this line prevents official releases)
    # should work again now, see #482 and obspy/obspy@b437f31
    if "-" not in line and "." not in line:
        version = "0.0.0.dev+0.g%s" % line
    else:
        parts = line.split('-', 1)
        version = parts[0]
        try:
            modifier = '+' if '.post' in version else '.post+'
            version += modifier + parts[1]
            if remote_tracking_branch is not None:
                version += '.' + remote_tracking_branch
        # IndexError means we are at a release version tag cleanly,
        # add nothing additional
        except IndexError:
            pass
    return version


def read_release_version():
    try:
        with io.open(VERSION_FILE, "rt") as fh:
            version = fh.readline()
        return version.strip()
    except IOError:
        return None


def write_release_version(version):
    with io.open(VERSION_FILE, "wb") as fh:
        fh.write(("%s\n" % version).encode('ascii', 'strict'))


def get_git_version(abbrev=10, dirty=True, append_remote_tracking_branch=True):
    # Read in the version that's currently in RELEASE-VERSION.
    release_version = read_release_version()

    # First try to get the current version using “git describe”.
    version = call_git_describe(
        abbrev, dirty=dirty,
        append_remote_tracking_branch=append_remote_tracking_branch)
    # If that doesn't work, fall back on the value that's in
    # RELEASE-VERSION.
    if version is None:
        version = release_version

    # If we still don't have anything, that's an error.
    if version is None:
        warnings.warn("ObsPy could not determine its version number. Make "
                      "sure it is properly installed. This for example "
                      "happens when installing from a zip archive "
                      "of the ObsPy repository which is not a supported way "
                      "of installing ObsPy.")
        return '0.0.0+archive'

    # pip uses its normalized version number (strict PEP440) instead of our
    # original version number, so we bow to pip and use the normalized version
    # number internally, too, to avoid discrepancies.
    version = _normalize_version(version)

    # If the current version is different from what's in the
    # RELEASE-VERSION file, update the file to be current.
    if version != release_version:
        write_release_version(version)

    # Finally, return the current version.
    return version


def _normalize_version(version):
    """
    Normalize version number string to adhere with PEP440 strictly.
    """
    pattern = (
        r'^[0-9]+?\.[0-9]+?\.[0-9]+?'
        r'((a|b|rc)[0-9]+?)?'
        r'(\.post[0-9]+?)?'
        r'(\.dev[0-9]+?)?$'
    )
    # we have a clean release version or another clean version
    # according to PEP 440
    if re.match(pattern, version):
        return version
    # we have an old-style version (i.e. a git describe string), prepare it for
    # the rest of clean up, i.e. put the '.post+' as separator for the local
    # version number part
    elif '.post' in version:
        version = re.sub(r'-', '+', version, count=1)
    elif re.match(r'^[0-9]+?\.[0-9]+?\.[0-9]+?-[0-9]+?-g[0-9a-z]+?$', version):
        version = re.sub(r'-', '.post+', version, count=1)
    # only adapt local version part right
    version = re.match(r'(.*?\+)(.*)', version)
    # no upper case letters
    local_version = version.group(2).lower()
    # only alphanumeric and "." in local part
    local_version = re.sub(r'[^A-Za-z0-9.]', r'.', local_version)
    version = version.group(1) + local_version
    # make sure there's a "0" after ".post"
    version = re.sub(r'\.post\+', r'.post0+', version)
    return version


if __name__ == "__main__":
    print(get_git_version())



# -*- coding: utf-8 -*-
"""
The obspy.io.segy Seismic Unix test suite.
"""
import io

import numpy as np

import obspy
from obspy.core.util import NamedTemporaryFile
from obspy.io.segy.segy import SEGYTraceReadingError, _read_su, iread_su
import pytest


class TestSU():
    """
    Test cases for SU reading and writing.

    Since the Seismic Unix format is a subset of the SEG Y file format a lot of
    the SEG Y tests cover certain aspects of the SU format and ensure that the
    SU implementation is working correctly.
    """
    def test_read_and_write_su(self, testdata):
        """
        Reading and writing a SU file should not change it.
        """
        file = testdata['1.su_first_trace']
        # Read the original file once.
        with open(file, 'rb') as f:
            org_data = f.read()
        with NamedTemporaryFile() as tf:
            outfile = tf.name
            # Read the SU file.
            su = _read_su(file)
            # Write it.
            su.write(outfile)
            with open(outfile, 'rb') as f:
                new_data = f.read()
        # Should be identical!
        assert org_data == new_data

    def test_enforcing_byteorders_while_reading(self, testdata):
        """
        Tests whether or not enforcing the byte order while reading and writing
        does something and works at all. Using the wrong byte order will most
        likely raise an Exception.
        """
        # This file is little endian.
        file = testdata['1.su_first_trace']
        # The following should both work.
        su = _read_su(file)
        assert su.endian == '<'
        su = _read_su(file, endian='<')
        assert su.endian == '<'
        # The following not because it will unpack the header and try to unpack
        # the number of data samples specified there which will of course not
        # correct.
        with pytest.raises(SEGYTraceReadingError):
            _read_su(file, endian='>')

    def test_reading_and_writing_different_byteorders(self, testdata):
        """
        Writing different byte orders should not change
        """
        # This file is little endian.
        file = testdata['1.su_first_trace']
        with NamedTemporaryFile() as tf:
            outfile = tf.name
            # The following should both work.
            su = _read_su(file)
            data = su.traces[0].data
            # Also read the original file.
            with open(file, 'rb') as f:
                org_data = f.read()
            assert su.endian == '<'
            # Write it little endian.
            su.write(outfile, endian='<')
            with open(outfile, 'rb') as f:
                new_data = f.read()
            assert org_data == new_data
            su2 = _read_su(outfile)
            assert su2.endian == '<'
            np.testing.assert_array_equal(data, su2.traces[0].data)
            # Write it big endian.
            su.write(outfile, endian='>')
            with open(outfile, 'rb') as f:
                new_data = f.read()
            assert not (org_data == new_data)
            su3 = _read_su(outfile)
        assert su3.endian == '>'
        np.testing.assert_array_equal(data, su3.traces[0].data)

    def test_unpacking_su_data(self, testdata):
        """
        Unpacks data and compares them to data unpacked by Madagascar.
        """
        # This file has the same data as 1.sgy_first_trace.
        file = testdata['1.su_first_trace']
        data_file = testdata['1.sgy_first_trace.npy']
        su = _read_su(file)
        data = su.traces[0].data
        # The data is written as integer so it is also converted to float32.
        correct_data = np.require(np.load(data_file).ravel(), np.float32)
        # Compare both.
        np.testing.assert_array_equal(correct_data, data)

    def test_read_bytes_io(self, testdata):
        """
        Tests reading from BytesIO instances.
        """
        # 1
        filename = testdata['1.su_first_trace']
        with open(filename, 'rb') as fp:
            data = fp.read()
        st = _read_su(io.BytesIO(data))
        assert len(st.traces[0].data) == 8000

    def test_iterative_reading(self, testdata):
        """
        Tests iterative reading.
        """
        # Read normally.
        filename = testdata['1.su_first_trace']
        st = obspy.read(filename, unpack_trace_headers=True)

        # Read iterative.
        ist = [_i for _i in iread_su(filename, unpack_headers=True)]

        del ist[0].stats.su.data_encoding

        assert st.traces == ist



# -*- coding: utf-8 -*-
import warnings

import pytest

from obspy.core.event import ResourceIdentifier, read_events
from obspy.core.utcdatetime import UTCDateTime
from obspy.core.util.base import NamedTemporaryFile
from obspy.io.pde.mchedr import _read_mchedr
from obspy.io.quakeml.core import _read_quakeml, _write_quakeml

# lxml < 2.3 seems not to ship with RelaxNG schema parser and namespace support
IS_RECENT_LXML = False
try:
    from lxml.etree import __version__
    version = float(__version__.rsplit('.', 1)[0])
    if version >= 2.3:
        IS_RECENT_LXML = True
except Exception:
    pass


class TestMchedr():
    """
    Test suite for obspy.mchedr
    """
    @pytest.fixture(autouse=True, scope="function")
    def setup(self, testdata):
        # read the mchedr file once for all
        self.catalog = _read_mchedr(testdata['mchedr.dat'])

    def test_catalog(self):
        assert len(self.catalog) == 1
        assert str(self.catalog) == \
            '''1 Event(s) in Catalog:
2012-01-01T05:27:55.980000Z | +31.456, +138.072 | 6.2  Mb'''

    def test_event(self):
        """
        Tests Event object.
        """
        event = self.catalog[0]
        assert event.resource_id == \
            ResourceIdentifier(
                id='quakeml:us.anss.org/event/20120101052755.98')
        # enums
        assert event.event_type is None
        assert event.event_type_certainty is None
        # comments
        assert len(event.comments) == 1
        c = event.comments
        assert c[0].text == 'MW 6.8 (WCMT), 6.8 (UCMT), 6.8 (GCMT). \
Felt (V) at Chiba; (IV) at Fussa, Kawasaki, Saitama, Tokyo, \
Yokohama and Yokosuka; (III) at Ebina, Zama and Zushi; (II) \
at Misawa and Narita, Honshu. Recorded (4 JMA) in Chiba, Fukushima, \
Gumma, Ibaraki, Kanagawa, Miyagi, Saitama, Tochigi and Tokyo.'
        # event descriptions
        assert len(event.event_descriptions) == 2
        d = event.event_descriptions
        assert d[0].text == 'SOUTHEAST OF HONSHU, JAPAN'
        assert d[0].type == 'region name'
        assert d[1].text == '211'
        assert d[1].type == 'Flinn-Engdahl region'
        # creation info
        assert event.creation_info is None

    def test_origin(self):
        """
        Tests Origin object.
        """
        assert len(self.catalog[0].origins) == 4
        origin = self.catalog[0].origins[0]
        assert origin.resource_id == \
            ResourceIdentifier(
                id='quakeml:us.anss.org/origin/20120101052755.98')
        assert origin.origin_type == 'hypocenter'
        assert origin.time == \
            UTCDateTime(2012, 1, 1, 5, 27, 55, 980000)
        assert origin.latitude == 31.456
        assert round(abs(origin.latitude_errors.uncertainty-0.0155), 3) == 0
        assert origin.longitude == 138.072
        assert round(abs(origin.longitude_errors.uncertainty-0.0173), 3) == 0
        assert origin.depth == 365300.0
        assert origin.depth_errors.uncertainty == 2700.0
        assert origin.depth_type == 'from location'
        assert origin.method_id is None
        assert origin.time_fixed is None
        assert origin.epicenter_fixed is None
        assert origin.earth_model_id == \
            ResourceIdentifier(
                id='quakeml:us.anss.org/earthmodel/ak135')
        assert origin.evaluation_mode is None
        assert origin.evaluation_status is None
        assert origin.origin_type == 'hypocenter'
        # composite times
        assert len(origin.composite_times) == 0
        # quality
        assert origin.quality.used_station_count == 628
        assert origin.quality.standard_error == 0.84
        assert origin.quality.azimuthal_gap == 10.8
        assert origin.quality.maximum_distance == 29.1
        assert origin.quality.minimum_distance == 2.22
        assert origin.quality.associated_phase_count == 52
        assert origin.quality.associated_station_count == 628
        assert origin.quality.depth_phase_count == 0
        assert origin.quality.secondary_azimuthal_gap is None
        assert origin.quality.ground_truth_level is None
        assert origin.quality.median_distance is None
        # comments
        assert len(origin.comments) == 0
        # creation info
        assert origin.creation_info.author is None
        assert origin.creation_info.agency_id == 'USGS-NEIC'
        assert origin.creation_info.author_uri is None
        assert origin.creation_info.agency_uri is None
        assert origin.creation_info.creation_time is None
        assert origin.creation_info.version is None
        # origin uncertainty
        u = origin.origin_uncertainty
        assert u.preferred_description == 'confidence ellipsoid'
        assert u.horizontal_uncertainty is None
        assert u.min_horizontal_uncertainty is None
        assert u.max_horizontal_uncertainty is None
        assert u.azimuth_max_horizontal_uncertainty is None
        # confidence ellipsoid
        c = u.confidence_ellipsoid
        assert c.semi_intermediate_axis_length == 2750.0
        # c.major_axis_rotation is computed during file reading:
        assert round(abs(c.major_axis_rotation-170.5), 3) == 0
        assert c.major_axis_plunge == 76.06
        assert c.semi_minor_axis_length == 2210.0
        assert c.semi_major_axis_length == 4220.0
        assert c.major_axis_azimuth == 292.79

    def test_magnitude(self):
        """
        Tests Magnitude object.
        """
        assert len(self.catalog[0].magnitudes) == 3
        mag = self.catalog[0].magnitudes[0]
        assert mag.resource_id == \
            ResourceIdentifier(
                id='quakeml:us.anss.org/magnitude/20120101052755.98/mb')
        assert mag.mag == 6.2
        assert mag.mag_errors.uncertainty is None
        assert mag.magnitude_type == 'Mb'
        assert mag.station_count == 294
        assert mag.evaluation_status is None
        # comments
        assert len(mag.comments) == 0
        # creation info
        assert mag.creation_info.author is None
        assert mag.creation_info.agency_id == 'USGS-NEIC'
        assert mag.creation_info.author_uri is None
        assert mag.creation_info.agency_uri is None
        assert mag.creation_info.creation_time is None
        assert mag.creation_info.version is None

    def test_stationmagnitude(self):
        """
        Tests StationMagnitude object.
        """
        assert len(self.catalog[0].station_magnitudes) == 19
        mag = self.catalog[0].station_magnitudes[0]
        assert mag.mag == 6.6
        assert mag.mag_errors.uncertainty is None
        assert mag.station_magnitude_type == 'Mb'
        assert mag.waveform_id.station_code == 'MDJ'
        assert mag.creation_info is None

    def test_amplitude(self):
        """
        Tests Amplitude object.
        """
        assert len(self.catalog[0].station_magnitudes) == 19
        amp = self.catalog[0].amplitudes[0]
        assert round(abs(amp.generic_amplitude-3.94502e-06), 7) == 0
        assert amp.type == 'AB'
        assert amp.period == 1.3
        assert amp.magnitude_hint == 'Mb'
        assert amp.waveform_id.station_code == 'MDJ'
        assert amp.creation_info is None

    def test_arrival(self):
        """
        Tests Arrival object.
        """
        assert len(self.catalog[0].origins[0].arrivals) == 52
        ar = self.catalog[0].origins[0].arrivals[0]
        assert ar.phase == 'Pn'
        assert ar.azimuth == 41.4
        assert ar.distance == 2.22
        assert ar.takeoff_angle is None
        assert ar.takeoff_angle_errors.uncertainty is None
        assert ar.time_residual == -1.9
        assert ar.horizontal_slowness_residual is None
        assert ar.backazimuth_residual is None
        assert ar.time_weight is None
        assert ar.horizontal_slowness_weight is None
        assert ar.backazimuth_weight is None
        assert ar.earth_model_id == \
            ResourceIdentifier('quakeml:us.anss.org/earthmodel/ak135')
        assert len(ar.comments) == 0

    def test_pick(self):
        """
        Tests Pick object.
        """
        assert len(self.catalog[0].picks) == 52
        pick = self.catalog[0].picks[0]
        assert pick.time == UTCDateTime(2012, 1, 1, 5, 28, 48, 180000)
        assert pick.time_errors.uncertainty is None
        assert pick.waveform_id.station_code == 'JHJ2'
        assert round(abs(pick.backazimuth--138.6), 7) == 0
        assert pick.onset == 'emergent'
        assert pick.phase_hint == 'Pn'
        assert pick.polarity is None
        assert pick.evaluation_mode is None
        assert pick.evaluation_status is None
        assert len(pick.comments) == 0

    def test_focalmechanism(self):
        """
        Tests FocalMechanism object.
        """
        assert len(self.catalog[0].focal_mechanisms) == 4
        fm = self.catalog[0].focal_mechanisms[0]
        assert fm.resource_id == \
            ResourceIdentifier(
                id='quakeml:us.anss.org/focalmechanism/'
                   '20120101052755.98/ucmt/mwc')
        # general
        assert fm.waveform_id == []
        assert fm.triggering_origin_id is None
        assert fm.azimuthal_gap is None
        assert fm.station_polarity_count is None
        assert fm.misfit is None
        assert fm.station_distribution_ratio is None
        assert fm.method_id == \
            ResourceIdentifier(
                id='quakeml:us.anss.org/methodID=CMT')
        # comments
        assert len(fm.comments) == 0
        # creation info
        assert fm.creation_info.author is None
        assert fm.creation_info.agency_id == 'UCMT'
        assert fm.creation_info.author_uri is None
        assert fm.creation_info.agency_uri is None
        assert fm.creation_info.creation_time is None
        assert fm.creation_info.version is None
        # nodalPlanes
        assert round(abs(fm.nodal_planes.nodal_plane_1.strike-5.0), 7) == 0
        assert round(abs(fm.nodal_planes.nodal_plane_1.dip-85.0), 7) == 0
        assert round(abs(fm.nodal_planes.nodal_plane_1.rake--76.0), 7) == 0
        assert round(abs(fm.nodal_planes.nodal_plane_2.strike-116.0), 7) == 0
        assert round(abs(fm.nodal_planes.nodal_plane_2.dip-15.0), 7) == 0
        assert round(abs(fm.nodal_planes.nodal_plane_2.rake--159.0), 7) == 0
        assert fm.nodal_planes.preferred_plane is None
        # principalAxes
        assert round(abs(fm.principal_axes.t_axis.azimuth-82.0), 7) == 0
        assert round(abs(fm.principal_axes.t_axis.plunge-38.0), 7) == 0
        assert round(abs(fm.principal_axes.t_axis.length-1.87e+19), 7) == 0
        assert round(abs(fm.principal_axes.p_axis.azimuth-290.0), 7) == 0
        assert round(abs(fm.principal_axes.p_axis.plunge-49.0), 7) == 0
        assert round(abs(fm.principal_axes.p_axis.length--1.87e+19), 7) == 0
        assert fm.principal_axes.n_axis.azimuth == 184
        assert fm.principal_axes.n_axis.plunge == 14
        assert fm.principal_axes.n_axis.length == 0.0
        # momentTensor
        mt = fm.moment_tensor
        assert mt.resource_id == \
            ResourceIdentifier(
                id='quakeml:us.anss.org/momenttensor/'
                   '20120101052755.98/ucmt/mwc')
        assert round(abs(mt.scalar_moment-1.9e+19), 7) == 0
        assert round(abs(mt.tensor.m_rr--3.4e+18), 7) == 0
        assert round(abs(mt.tensor.m_tt--8e+17), 7) == 0
        assert round(abs(mt.tensor.m_pp-4.2e+18), 7) == 0
        assert round(abs(mt.tensor.m_rt--1.9e+18), 7) == 0
        assert round(abs(mt.tensor.m_rp--1.77e+19), 7) == 0
        assert round(abs(mt.tensor.m_tp--4.2e+18), 7) == 0
        assert mt.clvd is None

    def test_write_quakeml(self):
        """
        Tests writing a QuakeML document.
        """
        with NamedTemporaryFile() as tf:
            _write_quakeml(self.catalog, tf, validate=IS_RECENT_LXML)
            # Read file again. Avoid the (legit) warning about the already used
            # resource identifiers.
            tf.seek(0)
            with warnings.catch_warnings(record=True):
                warnings.simplefilter("ignore")
                catalog2 = _read_quakeml(tf)
        assert len(catalog2), 1

    def test_read_events(self, testdata):
        """
        Tests reading an mchedr document via read_events.
        """
        filename = testdata['mchedr.dat']
        # Read file again. Avoid the (legit) warning about the already used
        # resource identifiers.
        with warnings.catch_warnings(record=True):
            warnings.simplefilter('always')
            catalog = read_events(filename)
            assert len(catalog), 1



# -*- coding: utf-8 -*-
"""
SeedLinkException.

Part of Python implementation of libslink of Chad Trabant and
JSeedLink of Anthony Lomax

:copyright:
    The ObsPy Development Team (devs@obspy.org) & Anthony Lomax
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""


class SeedLinkException(Exception):
    """
    SeedLink error.
    """
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)
